# -*- coding: utf-8 -*-
"""Provides the interface for the extraction of command line parameters
controlling the interconnection of *pydevrdc*.
"""
from __future__ import absolute_import
from __future__ import print_function

import sys
import os
import re
import glob
import argparse
import platform
import socket

try:
    import psutil
except:
    pass

from pythonids import ISSTR
from platformids import RTE, RTE_WIN32

from sourceinfo.fileinfo import getcaller_linenumber
from sourceinfo.objectinfo import getcaller_module_name

import filesysobjects
from filesysobjects.pathtools import expandpath, clearpath
from filesysobjects.userdata import gethome

from multiconf.apps.environ import ConfigEnviron

from jsondata.jsondata import JSONData

import rdbg
import rdbg.shared
from rdbg import PyRdbgError, RdbgNetworkError, callfromPyDevDbg
from rdbg.shared import _rdbgsrcmap_default, _rdbgsrcmapsep_default, _rdbglabel_default, \
    _rdbgfwd_default, _rdbg_default, _rdbgpath_default, _rdbgroot_default, _rdbgsub_default, \
    _rdbgtestflags, _rdbgtestflags_valid, pydevd_module

__author__ = 'Arno-Can Uestuensoez'
__license__ = "Artistic-License-2.0 + Forced-Fairplay-Constraints"
__copyright__ = "Copyright (C) 2010-2019 Arno-Can Uestuensoez" \
                " @Ingenieurbuero Arno-Can Uestuensoez"
__version__ = '0.2.23'
__uuid__ = '7f464182-b723-4e6e-869d-80a16b9cfaca'

__docformat__ = "restructuredtext en"

#
# The current values for external module access
#
_rdbgthis = False
_rdbgsrv = _rdbg_default
_rdbgfwd = _rdbgfwd_default
_rdbgroot = _rdbgroot_default
_rdbgsub = _rdbgsub_default
_rdbgpath = _rdbgpath_default
_rdbgsrcmap = _rdbgsrcmap_default
_rdbgsrcmapsep = _rdbgsrcmapsep_default
_rdbglabel = _rdbglabel_default

#
# static re
_reRDSAddr = re.compile(r"([^:]*):([^:]*)")

#
# RemoteDebugServer address
if RTE & RTE_WIN32:
    _reHostOnly = re.compile(r"""^[^:\\\\]+$""")
else:
    _reHostOnly = re.compile(r"""^[^:""" + os.sep + """]+$""")

_rePortOnly = re.compile(r"""^:[0-9]+$""")
_reServiceOnly = re.compile(r"""^:[^0-9].*$""")
_reHostPort = re.compile(r"""^([^:]+):([0-9]+)$""")
_reHostService = re.compile(r"""^([^:]+):([^0-9].*)$""")

# : Force or prohibit the read of environment variables, by default
# : used in priority order when present.
_rdbgenv = False


class RdbgCheckOptionsError(PyRdbgError):
    """Option parsing and the removal of *--rdbg-\** options.
    """
    pass


class RdbgScanPydevdError(PyRdbgError):
    pass


class RdbgSpace(object):
    """Common name space class as processing cache for the RDBG OptParser.
    The initiation process fetches and superposes defaults from various 
    sources - before the command line options. ::

       0. set hard-coded initial defaults
       1. add/superpose standard configuration file *rdbg.conf* 
          or selected custom file defaults
       2. add/superpose environment variable defaults, 
          requires activation option *--rdbg-env*
       3. fetch and add/superpose command line options

    """

    def __init__(self, argv=None, **kargs):
        """Initialize object.
        """
        self.appname = getcaller_module_name(4)
        if not argv:
            argv = sys.argv
        self._argv_org = argv[:]
        self._argv = argv
        self._known = []
        self.rdbgenv = _rdbgenv

        _auto = kargs.get('enableautoload') 
        _autopath = kargs.get('enableautopath') 

        # 0. add hard-coded defaults
        self.fetchhardcoded()

        # 1. add configuration file defaults
        if not _auto:
            self.fetchconfigfile()

        # 2. add environment variable defaults - requires activation option
        self.fetchenvironment()

        # 3. add configuration file defaults
        if _auto:
            self.fetchconfigfile(
                enableautoconf=True,
                enableautopath=_autopath
                )

        # 4. fetch and add command line options

    def fetchhardcoded(self):
        """Fetch and set hardcoded final defaults.
        """
        if rdbg.my_verb_self:
            sys.stdout.write(
                "VERB:%s:%d:Fetch hard-coded options\n" % (
                    __name__,
                    getcaller_linenumber(),
                    )
                )

        self.rdbg_srcmap = _rdbgsrcmap_default
        self.rdbg_srcmapsep = _rdbgsrcmapsep_default
        self.rdbg_label = _rdbglabel_default
        self.rdbg_forward = _rdbgfwd_default
        self.rdbg_testflags = _rdbgtestflags
        self.rdbg_srv = _rdbg_default
        self.rdbg_path = _rdbgpath_default
        self.rdbg_root = _rdbgroot_default
        self.rdbg_sub = _rdbgsub_default
        self.rdbg_get_stub = False

    def fetchenvironment(self):
        """Fetch and set data from environment variables.
        Enabled only when the option *--rdbg-env* is present.
        Sets *self.rdbg_env* to *True* when option is present.
        """
        if '--rdbg-env' not in sys.argv:
            return
        self.rdbg_env = True

        if rdbg.my_verb_self:
            sys.stdout.write(
                "VERB:%s:%d:Fetch environment variables\n" % (
                    __name__,
                    getcaller_linenumber(),
                    )
                )

        if os.environ.get('RDBGSRV'):
            self.rdbg_srv = os.environ.get('RDBGSRV')
        if os.environ.get('RDBGPATH'):
            self.rdbg_path = os.environ.get('RDBGPATH')
        if os.environ.get('RDBGROOT'):
            self.rdbg_root = os.environ.get('RDBGROOT')
        if os.environ.get('RDBGSUB'):
            self.rdbg_sub = os.environ.get('RDBGSUB')
        if os.environ.get('RDBGSRCMAP'):
            self.rdbg_srcmap = os.environ.get('RDBGSRCMAP')

    def fetchconfigfile(self, **kargs):
        """Fetch and set data from standard configuration file 
        or auto configuration file.
        
        Args:
            kargs:
                **enableautoconf**:
                    Activates the scan for headless auto 
                    configuation. ::

                       auto := (
                            True
                          | False
                       )

                    * *True*:
                        If no *--rdbg-\** command line options are present,
                        than scans for the file *rdbg-auto.conf/json*. When 
                        present loads the configuration and starts debugging,
                        else proceeds a normal startup.  
                        
                    * *False*:
                        Checks the command line for the option *--rdbg-config*,
                        when present loads the configuration file.

                **enableautopath**:
                    Adapts the search path.

        """
        _auto = kargs.get("enableautoconf", False)  # : switches to headless auto-configuration mode
        _autopath = kargs.get("enableautopath", None)  # : registers an alternate path for 'rdbg-auto.conf/json'

        # early check for load on demand only
        has_rdbg = False  # : rdbg options are present
        has_conf = None  # : configuration file is selected, considered only for standard startup

        for i in sys.argv:  # check command line for rdbg options
            
            if i.startswith('--rdbg') or i.startswith('-rdbg'):
                has_rdbg = True
            
            if i.startswith('--rdbg-config') or i.startswith('-rdbg-config'):
                has_conf = i
                break

        if _auto:
            if has_rdbg:
                raise PyRdbgError("auto-conf mode prohibit '--rdbg*' options: " + str(sys.argv))

            # proceed with auto configuration mode
            conf = 'rdbg-auto.conf'

            # set path for 'rdbg-auto.conf'
            if _autopath == None:
                _path = os.curdir 
            else:
                _path = _autopath

        else:
            _path = None  # default

            if has_conf:
                if '=' in has_conf:
                    conf = has_conf[has_conf.find('=') + 1:]
    
                else:
                    ix = sys.argv.index(has_conf)
                    try:
                        if sys.argv[ix + 1][0] != '-':
                            conf = sys.argv[ix + 1]
                        else:
                            conf = 'rdbg.conf'
                    except IndexError:
                        conf = 'rdbg.conf'

                # _path = os.path.abspath(os.path.dirname(conf))

            else:
                conf = None

        if conf and not os.path.isabs(conf):
            _sub = os.path.dirname(conf)
            if _sub:
                _path += os.sep + _sub

        if conf and conf not in ('help', '?',):
#             if not os.path.exists(conf):
#                 raise PyRdbgError(
#                     'Cannot load config\nmissing: %s\ncurdir:  %s' % (
#                     conf,
#                     os.path.abspath(os.path.curdir)
#                     ))
#             if not os.path.isfile(conf):
#                 raise PyRdbgError('Provided configuration is not a file:' + str(conf))

            pathobj = filesysobjects.configdata.ConfigPath(
                prepend=[_path],
#                prepend=[os.path.abspath(os.path.dirname(conf))],
                filext=['cfg', 'conf', 'json', ]
                )

            if rdbg.my_dbg_self:
                sys.stderr.write(
                    "RDBG:%s:%d:pathobj = %s\n" % (
                    __name__, getcaller_linenumber(), str(pathobj)))

            _cp = ConfigEnviron(
                conf,
                pathobj=pathobj,
                copy='deepcopy'
                )

            if rdbg.my_verb_self:
                sys.stdout.write(
                    "VERB:%s:%d:Fetch configuration file: %s\n" % (
                        __name__,
                        getcaller_linenumber(),
                        _cp.get_config_filepath(conf)
                        )
                    )

            _cdata = _cp.readout_data()

            print("4TEST:post: " + str(rdbg.my_dbg_self))

            print("4TEST:#------------------------------------------")
            
#             for k,v in _cdata['environ'].items():
#                 print("4TEST:_cdata[environ]:%s = %s" % (str(k), str(v)))
# 
#             print("4TEST:#------------------------------------------")
# 
#             for k,v in _cdata.items():
#                 print("4TEST:_cdata:%s = %s" % (str(k), str(v)))
# 
#             print("4TEST:#------------------------------------------")
#             
#             
#             for k,v in _cdata['environ'].items():
#                 print("4TEST:_cdata[environ]:JSONDATA:%s" % (str(k),))
#                 print("4TEST:_cdata[environ]:JSONDATA:%s" % (str(JSONData({k: v}))))
#                 print("4TEST:_cdata[environ]:JSONDATA:%s = %s" % (str(JSONData(k)), str(JSONData(v))))
# 
#             print("4TEST:#------------------------------------------")

            l = len(_cdata)
            _sub = {}
            idx = 0
            for x in _cdata:
                _sub[x] = _cdata[x]

            for x in _cdata['environ']:
                _sub['environ'][x] = _cdata['environ'][x]

#                         print("4TEST:#A***---")
#                         print("4TEST:_cdata:JSONDATA:sub: %d - %d %s = %s" % (l, idx, str(x), str(_sub[x])))
#                         print("4TEST:_cdata:JSONDATA:sub: %d - %d type: %s = %s" % (l, idx, str(type(x)), str(type(_sub[x]))))
#                         print("4TEST:_cdata:JSONDATA:sub: %d - %d %s => %s" % (l, idx, str(JSONData(x)), str(JSONData(_sub))))
#                         idx += 1
#                         print("4TEST:#B***---")

            print("4TEST:#Z***---")
            print("4TEST:_cdata:JSONDATA:sub: %d - %d %s" % (l, idx, str(JSONData(_sub))))
            print("4TEST:#ZZ***---")

            for k, v in _cdata.items():
                print("4TEST:_cdata:JSONDATA:%s" % (str(k)))
                print("4TEST:_cdata:JSONDATA:%s" % (str(v)))
                print("4TEST:#---")
                try:
                    print("4TEST:_cdata:JSONDATA:%s = %s" % (str(JSONData(k)), str(JSONData(v))))
                    print("4TEST:# 343")
                    print("4TEST:#---")
                except:
                    
                    for x in v:
                        print("4TEST:############")
                        try:
                            print("4TEST:s:x = " + str(x) + "  =  " + str(v[x]))
                        except:
                            print("4TEST:t:x = " + str(x))
                        print("4TEST:#***********")
                    
                    l = len(v)
                    _sub = {}
                    idx = 0
                    for x in v:
                        _sub[x] = v[x]
                        print("4TEST:#A***---")
                        print("4TEST:_cdata:JSONDATA:sub: %d - %d %s = %s" % (l, idx, str(x), str(_sub[x])))
                        print("4TEST:_cdata:JSONDATA:sub: %d - %d type: %s = %s" % (l, idx, str(type(x)), str(type(_sub[x]))))
                        print("4TEST:_cdata:JSONDATA:sub: %d - %d %s => %s" % (l, idx, str(JSONData(x)), str(JSONData(_sub))))
                        idx += 1
                        print("4TEST:#B***---")
                    
                    raise
            print("4TEST:#------------------------------------------")
            print("4TEST:x:#------------------------------------------")

            if rdbg.my_dbg_self > 2:
                sys.stderr.write(
                    "RDBG:%s:%d:loaded config:\n   data = %s\n" % (
                        __name__, getcaller_linenumber(),
                        str(JSONData(_cdata))
                    )
                )
            print("4TEST:after: " + str(rdbg.my_dbg_self))

            if _cdata.get('headless'):
                setattr(self, 'headless', _cdata['headless'])

            if _cdata.get('ignore'):
                # not really required
                setattr(self, 'ignore', _cdata['ignore'])

            if _cdata.get('srcmap'):
                # list of source maps
                v = ''.join(_cdata['srcmap'])
                v = v.split('\n')
                setattr(self, 'srcmap', v)

            if _cdata.get('pydev'):
                if _cdata['pydev'].get('host'):
                    setattr(self, 'host', _cdata['pydev']['host'])
                if _cdata['pydev'].get('port'):
                    setattr(self, 'port', _cdata['pydev']['port'])
                if _cdata['pydev'].get('rdbg_fwd'):
                    setattr(self, 'rdbg_fwd', _cdata['pydev']['rdbg_fwd'])
                if _cdata['pydev'].get('rdbg_label'):
                    setattr(self, 'rdbg_label', _cdata['pydev']['rdbg_label'])
                if _cdata['pydev'].get('rdbg_path'):
                    setattr(self, 'rdbg_path', _cdata['pydev']['rdbg_path'])
                if _cdata['pydev'].get('rdbg_root'):
                    setattr(self, 'rdbg_root', _cdata['pydev']['rdbg_root'])
                if _cdata['pydev'].get('rdbg_sub'):
                    setattr(self, 'rdbg_sub', _cdata['pydev']['rdbg_sub'])

        return True

    def __repr__(self):
        """Standard *repr()* of debug session data.
        """
        res = '{'
        for k, v, in sorted(self.__dict__.items()):
            _frmstr = "'%s': '%s', "
            if k in ('_argv', '_argv_org',):
                continue
            if k in ('rdbg_srcmap',):
                if v and type(v) is not list:
                    v = v.split(os.linesep)
                try:
                    # create final JSON format
                    v = ','.join(v)
                    v = '[' + v + ']'
                except:
                    pass
                res += _frmstr % (str(k), str(v))
                continue
            res += _frmstr % (str(k), str(v))
        res += '}'
        return res

    def __str__(self, **kw):
        """Standard *str()* of debug session data.
        """
        _indent = kw.get('indent', 0)
        res = ''
        _frmstr = "\n" + ' ' * _indent + "%-" + str(20) + "s: %s"
        for k, v, in sorted(self.__dict__.items()):
            if k in ('_argv', '_argv_org',):
                continue
            elif k in ('rdbg_srcmap',):
                if v and type(v) is not list:
                    v = v.split(os.linesep)
                try:
                    # create final JSON format
                    v = ','.join(v)
                    v = '[' + v + ']'
                except:
                    pass
                res += _frmstr % (str(k), str(v))
                continue

            res += _frmstr % (str(k), str(v))
        return res

    def get_JSON_bin(self):
        """Provides structured session data for the currrent connection
        in JSON format. ::

           result := {
             'appname': '',
             'rdbg': '',
             'rdbg_Version': '',
             'rdbg_config': '',
             'rdbg_env': '',
             'rdbg_forward': '',
             'rdbg_get_stub': '',
             'rdbg_get_stub_mod': '',
             'rdbg_h': '',
             'rdbg_help': '',
             'rdbg_keepargv': '',
             'rdbg_label': '',
             'rdbg_path': '',
             'rdbg_root': '',
             'rdbg_self': '',
             'rdbg_self': '',
             'rdbg_show': '',
             'rdbg_srcmap': '',
             'rdbg_srv': '',
             'rdbg_sub': '',
             'rdbg_testflags': '',
             'rdbg_version': '',
          }

        """
        j = {}
        j['appname'] = self.appname
        j['rdbg'] = self.rdbg
        j['rdbg_Version'] = self.rdbg_Version
        j['rdbg_config'] = self.rdbg_config
        j['rdbg_env'] = self.rdbg_env
        j['rdbg_forward'] = self.rdbg_forward
        j['rdbg_get_stub'] = self.rdbg_get_stub
        j['rdbg_get_stub_mod'] = self.rdbg_get_stub_mod
        j['rdbg_h'] = self.rdbg_h
        j['rdbg_help'] = self.rdbg_help
        j['rdbg_keep_argv'] = self.rdbg_keep_argv
        j['rdbg_label'] = self.rdbg_label
        j['rdbg_path'] = self.rdbg_path
        j['rdbg_root'] = self.rdbg_root
        j['rdbg_self'] = self.rdbg_self
        j['rdbg_show'] = self.rdbg_show
        j['rdbg_srcmap'] = self.rdbg_srcmap
#            j['rdbg_srv'] =  self.rdbg_srv
        j['rdbg_srv'] = self.rdbg
        j['rdbg_sub'] = self.rdbg_sub
        j['rdbg_testflags'] = self.rdbg_testflags
        j['rdbg_version'] = self.rdbg_version
        j['rdbgenv'] = self.rdbgenv
        return j

    def get_items(self):
        """Provides a list of flat str-value tuples of session data
        for the currrent connection.
        ::

           result := [
              ('appname', ''),
              ('rdbg', ''),
              ('rdbg_Version', ''),
              ('rdbg_config', ''),
              ('rdbg_env', ''),
              ('rdbg_forward', ''),
              ('rdbg_get_stub', ''),
              ('rdbg_get_stub_mod', ''),
              ('rdbg_h', ''),
              ('rdbg_help', ''),
              ('rdbg_keepargv', ''),
              ('rdbg_label', ''),
              ('rdbg_path', ''),
              ('rdbg_root', ''),
              ('rdbg_self', ''),
              ('rdbg_self', ''),
              ('rdbg_show', ''),
              ('rdbg_srcmap', ''),
              ('rdbg_srv', ''),
              ('rdbg_sub', ''),
              ('rdbg_testflags', ''),
              ('rdbg_version', ''),
           ]

        """
        return [
            ('appname', self.appname),
            ('rdbg', self.rdbg),
            ('rdbg_Version', self.rdbg_Version),
            ('rdbg_config', self.rdbg_config),
            ('rdbg_env', self.rdbg_env),
            ('rdbg_forward', self.rdbg_forward),
            ('rdbg_get_stub', self.rdbg_get_stub),
            ('rdbg_get_stub_mod', self.rdbg_get_stub_mod),
            ('rdbg_h', self.rdbg_h),
            ('rdbg_help', self.rdbg_help),
            ('rdbg_keepargv', self.rdbg_keep_argv),
            ('rdbg_label', self.rdbg_label),
            ('rdbg_path', self.rdbg_path),
            ('rdbg_root', self.rdbg_root),
            ('rdbg_self', self.rdbg_self),
            ('rdbg_show', self.rdbg_show),
            ('rdbg_srcmap', self.rdbg_srcmap),
#                ('rdbg_srv', self.rdbg_srv),
            ('rdbg_srv', self.rdbg),
            ('rdbg_sub', self.rdbg_sub),
            ('rdbg_testflags', self.rdbg_testflags),
            ('rdbg_version', self.rdbg_version),
        ]


def fetch_rdbg_opts(argv=None, **kargs):
    """Reads - pops - *rdbg* options from *argv* by removing.

    Args:
        **argv**:
            The 'argv' to be parsed.

        kargs:

            **keepargv**:
                Suppress removal of known options from 'argv'.

                *default := False*

            **optspace**:
                Name space container of type 'RdbgSpace' to be used.

                *default := local-temp*

    Returns:
        The list of remaining options/arguments.

    Raises:
        pass-through
        
    """
    appname = kargs.get('appname', 'rdbg')
    keepargv = kargs.get('keepargv', False)
    optspace = kargs.get('optspace')
    if optspace == None:
        optspace = RdbgSpace()

    class FormatTextRaw(argparse.HelpFormatter):
        """Formatter for help."""

        def _split_lines(self, text, width):
            """Customize help format."""
            if text.startswith('@R:'):
                return text[3:].splitlines()
            return argparse.HelpFormatter._split_lines(self, text, width)

    class ActionExt(argparse.Action):
        """Extends the 'argparse.Action'
        """

        def __init__(self, *args, **kwargs):
            super(ActionExt, self).__init__(*args, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            """Adds context help on each option.
            """
            if rdbg.my_dbg_self > 2:
                if values:
                    sys.stderr.write(
                        "RDBG:%s:%d:CLI:option: %s=%s\n" % (
                            __name__,
                            getcaller_linenumber(),
                            str(option_string),
                            str(values)
                            )
                        )
                else:
                    sys.stderr.write(
                        "RDBG:%s:%d:CLI:option: %s\n" % (
                            __name__,
                            getcaller_linenumber(),
                            str(option_string)
                            )
                        )

            if values and (values in ("help", "?") or type(values) is list and values[0] in ("help", "?")):
                form = FormatTextRaw('rdbg')
                # print('\n' + str(sys.argv[-2]))
                if self.__doc__:
                    print('\n  ' + 
                          '\n  '.join(FormatTextRaw._split_lines(
                              form, "@R:" + self.__doc__, 80)) + '\n')
                else:
                    print('\n  ' + 
                          '\n  '.join(FormatTextRaw._split_lines(
                              form, "No help.", 80)) + '\n')
                sys.exit(0)

            self.call(parser, namespace, values, option_string)

        def get_help_text(self):
            form = FormatTextRaw('rdbg')
            # print('\n' + str(sys.argv[-2]))
            if self.__doc__:
                return '\n  ' + '\n  '.join(FormatTextRaw._split_lines(
                    form, "@R:" + self.__doc__, 80)) + '\n'
            else:
                return '\n  ' + '\n  '.join(FormatTextRaw._split_lines(
                    form, "No help.", 80)) + '\n'

        def help_option(self):
            print(self.get_help_text())
            sys.exit(0)

        def remove_option(self, parser, namespace, values, option_string=None):
            _ix = -1
            _iv = -1
            _eq = False

            try:
                _ix = namespace._argv.index(option_string)
                if type(values) is list:
                    _iv = namespace._argv.index(values[0])
                else:
                    _iv = namespace._argv.index(values)
            except Exception:
                try:
                    if type(values) is list:
                        _ix = namespace._argv.index(option_string + "=" + values[0])
                    else:
                        _ix = namespace._argv.index(option_string + "=" + str(values))
                    _eq = True
                except Exception:
                    pass

            if _iv > 0 and not keepargv:
                _ai = namespace._argv.pop(_iv)

            if _ix > 0 and not keepargv:
                _ai = namespace._argv.pop(_ix)

            return

    class OptActionConfig(ActionExt):
        """Reads the selected configuration file.

  --rdbg-config[=<configuration-file-path-name>]

    configuration-file-path-name := (<abs-file>|<rel-file>)
    abs-file := "absolute file path name"
    rel-file := "relative file path name"

    default  := rdbg.conf

For security reasons only literal file path names 
are supported, without search operations.

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values is None:
                if not os.path.exists('rdbg.conf'):
                    raise PyRdbgError('Default-Config "rdbg.conf" does not exist in:' + str(os.path.abspath(os.path.curdir)))
                namespace.rdbg_config = 'rdbg.conf'

            else:
                if not os.path.exists(values):
                    raise PyRdbgError('Config does not exist:' + str(values))
                self.rdbg_config = values

    class OptActionDBG(ActionExt):
        """Set the debugging of RDBG itself

  --rdbg-self[=#level]

    level: "integer value of the debug level"

Each repetition without a value increments the level by 1.

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionDBG, self).__init__(
                option_strings, dest, nargs, **kwargs)

            self.reset_val = False

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values and values.isdigit():
                # reset early fetch for the case of '=' based value separators 
                rdbg.my_dbg_self = int(values)
                self.reset_val = True

            elif self.reset_val:
                rdbg.my_dbg_self += 1

            # from early fetch
            setattr(namespace, self.dest, rdbg.my_dbg_self)

    class OptActionPrintOptions(ActionExt):
        """Print provided command line options.

  --rdbg-options [<format> [<proceed>]]
  --rdbg-options=<format>

    format  :=  (json | repr | str)
    proceed :=  (break | continue | cont)

    json:       "JSON format"
    repr:       "Python repr-format"
    str:        "string"

    break :     "exits after print"
    continue :  "continues after print"
    cont :      "continues after print"

   default := json break

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionPrintOptions, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            namespace.rdbg_options = ['json', 'break', ]
            if not values:
                pass
            else:
                try:
                    namespace.rdbg_options[0] = values[0].lower()
                    namespace.rdbg_options[1] = values[1].lower()
                except IndexError:
                    pass

            if namespace.rdbg_options[0] not in ('json', 'repr', 'str',):
                raise PyRdbgError("invalid value: --rdbg-options=" + str(namespace.rdbg_options[0]))
            elif namespace.rdbg_options[1] not in ('break', 'continue', 'cont',):
                raise PyRdbgError("invalid value: --rdbg-options=" + str(namespace.rdbg_options[1]))

    class OptActionEnv(ActionExt):
        """Enables the readout of RDBG environment variables:

  --rdbg-env := (True|False)

Currently: RDBGROOT, RDBGSUB, RDBGSRV

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values is None:
                setattr(namespace, self.dest, True)

            else:
                if values.lower() in (
                        'on',
                        'true',
                        '1',
                        True,
                        1,):
                    setattr(namespace, self.dest, True)

                elif values.lower() in (
                        'off',
                        'false',
                        '0',
                        False,
                        0,):
                    setattr(namespace, self.dest, False)

                else:
                    raise RdbgCheckOptionsError(
                        "Unknown value:" + str(values))

    class OptActionFWD(ActionExt):
        """Forward the '--rdbg' option to subprocesses for nested debugging
of process chains.

  --rdbg-forward=(<forwarding-levels>|all|label)

    <forwarding-levels>:
        Number of levels to be forwarded, 0==None.

    all:
        all nested subprocesses

    label:
        An arbitrary label defined at initialization of the
        debug instance. Debugging is enabled when these match.

    default:=0: No forwarding.

        """

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if not values:  # default
                setattr(namespace, self.dest, _rdbgfwd_default)
                return

            elif type(values) is int and values > 0:  # integer
                setattr(namespace, self.dest, (values))

            elif values.isdigit() and int(values) > 0:  # integer
                setattr(namespace, self.dest, int(values))

            elif values == 'all':  # all levels
                setattr(namespace, self.dest, values)

    class OptActionGetStub(ActionExt):
        """Displays the resulting path to the current stub module:

  --rdbg-get-stub

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            setattr(namespace, self.dest, True)

    class OptActionGetStubMod(ActionExt):
        """Displays the current resulting stub module:

  --rdbg-get-stub-mod

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            setattr(namespace, self.dest, True)

    class OptActionRDBGS(ActionExt):
        """Activate remote debug, the host and port number of
the server process could be changed.

  --rdbg [host][:port]

    default := localhost:5678

Examples:

  --rdbg
  --rdbg localhost
  --rdbg :5678
  --rdbg localhost:5678
  --rdbg 127.0.0.1:5678

Current support for IPv4 only.

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionRDBGS, self).__init__(option_strings, dest, nargs,
                                              **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if not values:
                setattr(namespace, self.dest, _rdbg_default)
                return

            elif _reHostOnly.match(values):
                setattr(namespace, self.dest, values + ":5678")

            elif _rePortOnly.match(values):
                setattr(namespace, self.dest, "localhost" + values)

            elif _reServiceOnly.match(values):
                setattr(namespace, self.dest, "localhost" + values)

            elif _reHostPort.match(values):
                setattr(namespace, self.dest, values)

            elif _reHostService.match(values):
                setattr(namespace, self.dest, values)

            else:  # no valid argument
                setattr(namespace, self.dest, _rdbg_default)
                return

    class OptActionKeepArgv(ActionExt):
        """Keep argv with contained RDBG options, else these
are removed before passing the control to application.

  --rdbg-keep-argv

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionKeepArgv, self).__init__(option_strings, dest, nargs,
                                              **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            if not values:
                setattr(namespace, self._rdbg_keep_argv, False)
                return
            setattr(namespace, self._rdbg_keep_argv, True)

    class OptActionPATH(ActionExt):
        """Set the default path directory for scanning for the
modules 'pydevd.py', and 'pydev_file_utils.py'.
This option takes precedence over the options '--rdbg-root'
and 'rdbg-sub'.

  --rdbg-path=(
         <pathforscan>
       | <FQDN>
       | <path-glob>
       | False
       )

    FQDN        := "fully qualified name"
    path-glob   := "glob for path, see 'filesysobjects'"
    pathforscan := "the path to the *pysrc* directory"
    False       := "suppress default lookup of installed standalone package 'pydevd'"

    default     := "try first 'import pydevd'"

For the directory path name optional wildcard and regular
expressions are supported based on 'glob' and 're', see

   'filesysobjects.findrelpath_in_searchpath'
   'scan_for_pydevd'

This option supports repetition, the paths are searched in the provided order.

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values in (False,):
                _v = False

            elif values in (None,):
                _v = _rdbgpath_default

            else:
                _v = values

            vx = getattr(namespace, self.dest)
            vx.append(_v)

    class OptActionLabel(ActionExt):
        """Set a label for the current process, could be used for
later addressing by *--rdbg-forward*.

  --rdbg-label <process-label>

"""

#         def __init__(self, option_strings, dest, nargs=None, **kwargs):
#             super(OptActionLabel, self).__init__(option_strings, dest, nargs,
#                                               **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            self.rdbg_label = values[0]

    class OptActionROOT(ActionExt):
        """Set the default root directory for scanning for the modules 'pydevd.py',
and 'pydev_file_utils.py'.

  --rdbg-root <root-dir>

With the search schemes:

  0. path := <root-dir>/pydevd.py
  1. path := <root-dir>/plugins/<sub-dir>

For the directory path names 'root-dir' and 'sub-dir' wildcard and
regular expressions based on 'glob' and 're' are supported,
see '--subdir' and

  'filesysobjects.findrelpath_in_searchpath'
  'scan_for_pydevd'

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)
            if not values:
                setattr(namespace, self.dest, os.pathsep.join(_rdbgroot_default))
                return

            else:
                setattr(namespace, self.dest, os.pathsep.join(values))

    class OptActionSUB(ActionExt):
        """Sets the default sub-directory for the scanning of the modules 'pydevd.py',
and 'pydev_file_utils.py'.

  --rdbg-sub  <sub-dir>

With the scheme:

  path := <root-dir>/plugins/<sub-dir>

For the director path names 'root-dir' and 'sub-dir' wildcard and regular
expressions based on 'glob' and 're' are supported, see

  'filesysobjects.findrelpath_in_searchpath'
  'scan_for_pydevd'

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if not values:
                setattr(namespace, self.dest, _rdbgsub_default)

            else:
                setattr(namespace, self.dest, values)

    class OptActionTFLAG(ActionExt):
        """Test flags for the debugging of RDBG itself,
could be repeated for setting multiple flags.

  --rdbg-testflags :=(
        ignorePydevd          # ignores loaded pydevd.py, for search test only
      | ignorePydevdSysPath   # ignores sys.path, for search test only
      | ignorePydevdCallParam # ignores call parameters
    )

"""

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values[0] in _rdbgtestflags_valid:
                if not namespace.rdbg_testflags:
                    namespace.rdbg_testflags = values
                else:
                    namespace.rdbg_testflags.extend(values)
            else:
                raise RdbgCheckOptionsError(
                    "Unknown value " + str(option_string) + "=" + 
                    str(values))

    class OptActionSHOW(ActionExt):
        """\nShows basic status with details.

  --rdbg-show=<scope>
  --rdbg-show <scope>[ <proceed>]

    scope := ('all'|'load'|'opt'|'rdbg'|'rt'|'python')
    load:    basic systems resource status
    opt:     resulting settings from options
    proc:    process id
    python:  Python runtime environment
    rdbg:    rdbg setup parameter values
    rt:      runtime platform
    all:     all

    proceed := 'cont'
      cont:   displays immediately when the option is matched,
              and continues to proceed

Exist by default after display.

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionSHOW, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            if not values:
                namespace.rdbg_show = ['all']
            elif values and values[0].lower() not in ('all', 'load', 'opt', 'rdbg', 'rt', 'python'):
                raise PyRdbgError("Unknown value: --rdbg-show=" + str(values))
            else:
                namespace.rdbg_show = values
                self.remove_option(parser, namespace, values, option_string)

        @classmethod
        def show_opt(cls, namespace, values):

            _osu = platform.uname()
            _dist, _distver, _x = platform.dist()

            _width = 18
            _fstr = "  %-" + str(_width) + "s: %s\n"

            if values is None:
                values = 'all'

            print("")
            if values[0].lower() in ('opt', 'all',):
                print(
                    "Current Debug Options\n" + 
                    "=" * 21 + 
                    "\n" + 
                    "rdbgoptions:\n" + 
                    _fstr % ("rdbgthis(0)", str(_rdbgthis)) + 
                    _fstr % ("rdbg(1)", str(namespace.rdbg)) + 
                    _fstr % ("rdbgfwd(2)", str(namespace.rdbg_forward)) + 
                    _fstr % ("rdbgroot(3)", str(namespace.rdbg_root)) + 
                    _fstr % ("rdbgsub(4)", str(namespace.rdbg_sub)) + 
                    _fstr % ("rdbgpath(5)", str(namespace.rdbg_path)) + 
                    _fstr % ("rdbgsrcmap(6)", str(namespace.rdbg_srcmap)) + 
                    _fstr % ("rdbglabel(7)", str(namespace.rdbg_label)) + 
                    "\n"
                )

            if values[0].lower() in ('rdbg', 'all',):
                import rdbg.start

                print(
                    "Current Setup Parameters of pyRDBG\n" + 
                    "=" * 34 + 
                    "\n" + 
                    "app:\n" + 
                    _fstr % ("appname", str(namespace.appname)) + 
                    "\n" + 
                    "pyrdbg:\n" + 
                    _fstr % ("id", str(__uuid__)) + 
                    _fstr % ("pyrdbg", str(rdbg.__version__)) + 
                    _fstr % ("rdbg.checkrdbg", str(__version__)) + 
                    _fstr % ("rdbg.pydevrdc", str(rdbg.pydevrdc.__version__)) + 
                    _fstr % ("rdbg.start", str(rdbg.start.__version__)) + 
                    "\n" + 
                    _fstr % ("rdbg-label", str(namespace.rdbg_label)) + 
                    _fstr % ("rdbg", str(namespace.rdbg)) + 
                    _fstr % ("srv", str(_rdbgsrv)) + 
                    _fstr % ("rdbg-forward", str(namespace.rdbg_forward)) + 
                    _fstr % ("this", str(_rdbgthis)) + 
                    "\n" + 
                    _fstr % ("rdbg-root", str(namespace.rdbg_root)) + 
                    _fstr % ("rdbg-sub", str(namespace.rdbg_sub)) + 
                    _fstr % ("rdbg-path", str(namespace.rdbg_path)) + 
                    "\n" + 
                    _fstr % ("pydevd.py", str(rdbg.__version__)) + 
                    "\n" + 
                    _fstr % ("rdbg-srcmap", str(namespace.rdbg_srcmap)) + 
                    "\n" + 
                    _fstr % ("rdbg-keep-argv", str(namespace.rdbg_keep_argv)) + 
                    _fstr % ("rdbg-env", str(namespace.rdbg_env)) + 
                    "\n"
                )

            if values[0].lower() in ('rt', 'all',):
                print(
                    "\n" + 
                    "Runtime Platform\n" + 
                    "="*16 + 
                    "\n" + 
                    "running python:\n" + 
                    _fstr % ("version" , str(platform.python_version())) + 
                    _fstr % ("compiler", str(platform.python_compiler())) + 
                    _fstr % ("build", str(platform.python_build())) + 
                    "\n" + 
                    "node:\n" + 
                    _fstr % ("login", str(os.getlogin())) + 
                    _fstr % ("host", str(platform.node())) + 
                    _fstr % ("ipv4", str(socket.gethostbyname(socket.gethostname())))
                )
                try:
                    print(_fstr % ("ipv6", str(socket.getaddrinfo(socket.gethostname(), 0, socket.AF_INET6))))
                except:
                    pass

                print(
                    "\n" + 
                    _fstr % ("arch", str(platform.machine())) + 
                    "\n" + 
                    _fstr % ("os", str(_osu[0])) + 
                    _fstr % ("os-rel", str(_osu[2])) + 
                    _fstr % ("dist", str(_dist)) + 
                    _fstr % ("dist-rel", str(_distver))
                )

            if values[0].lower() in ('python', 'all',):
                print(
                    "\n" + 
                    "Python Environment\n" + 
                    "="*18 + 
                    "\n" + 
                    "node:\n" + 
                    _fstr % ("login", str(os.getlogin())) + 
                    _fstr % ("host", str(platform.node())) + 
                    _fstr % ("arch", str(platform.machine())) + 
                    _fstr % ("os", str(_osu[0])) + 
                    _fstr % ("os-rel", str(_osu[2])) + 
                    _fstr % ("dist", str(_dist)) + 
                    _fstr % ("dist-rel", str(_distver)) + 
                    "\n" + 
                    "running python:\n" + 
                    _fstr % ("executable" , str(sys.executable)) + 
                    _fstr % ("version" , str(platform.python_version())) + 
                    _fstr % ("compiler", str(platform.python_compiler())) + 
                    _fstr % ("build", str(platform.python_build())) + 
                    _fstr % ("'os' path", str(os.__file__)) + 
                    _fstr % ("'checkrdbg' path", str(__file__))
                )

                print(_fstr % ("sys.path", "["))
                i = 0
                for s in sys.path:
                    sys.stdout.write(_fstr % ("  [%2d]" % (i), "  " + str(s)))
                    i += 1
                print(_fstr % ("", "]"))

            if values[0].lower() in ('load', 'all',):
                print(
                    "\n" + 
                    "System Load\n" + 
                    "="*11 + 
                    "\n" + 
                    _fstr % ("host", str(platform.node())) + 
                    _fstr % ("ipv4", str(socket.gethostbyname(socket.gethostname())))
                )

                try:
                    print(_fstr % ("ipv6", str(socket.getaddrinfo(socket.gethostname(), 0, socket.AF_INET6))))
                except:
                    pass

                print(
                    "\n" + 
                    "Processor\n" + 
                    _fstr % ("Arch", str(platform.processor()))
                )

                try:
                    print(_fstr % ("Cores/Threads", str(psutil.NUM_CPUS)))  # @UndefinedVariable
                except:
                    try:
                        print(_fstr % ("Cores/Threads", str(psutil.cpu_count())))  # @UndefinedVariable
                    except:
                        pass

                print("")

                _fstr2 = "  %-" + str(_width) + "s: %s%%"
                print(_fstr2 % ("cpu", str(psutil.cpu_percent())))

                _fstr3 = "  %-" + str(_width) + "s: %s%% of %sMi"
                try:
                    vmem = psutil.virtual_memory()
                    print(_fstr3 % ("phymem", str(vmem.percent), str(vmem.total / 1000000)))
                except:
                    pass

                try:
                    smem = psutil.swap_memory()
                    print(_fstr3 % ("swap", str(smem.percent), str(smem.total / 1000000)))
                except:
                    pass

                try:
                    print(_fstr % ("running procs", str(len(psutil.get_pid_list()))))  # @UndefinedVariable
                except:
                    try:
                        print(_fstr % ("running procs", str(len(psutil.pids()))))  # @UndefinedVariable
                    except:
                        pass

            if values[0].lower() in ('proc', 'all',):
                print(
                    "\n" + 
                    "Process ID\n" + 
                    "="*11 + 
                    "\n" + 
                    "node:\n" + 
                    _fstr % ("host", str(platform.node())) + 
                    _fstr % ("ipv4", str(socket.gethostbyname(socket.gethostname())))
                )
                try:
                    print(_fstr % ("ipv6", str(socket.getaddrinfo(socket.gethostname(), 0, socket.AF_INET6))))
                except:
                    pass
                print(
                    "\n" + 
                    "proc:\n" + 
                    _fstr % ("pid", str(os.getpid())) + 
                    _fstr % ("ppid", str(os.getppid()))
                )
            if values[0].lower() not in ('proc', 'opt', 'rdbg', 'rt', 'load', 'all', 'python'):
                raise PyRdbgError("Unknown value --rdbg-show=%s\n%s" % (str(values), cls.get_help_text()))

            print("")
            if not values[-1] or len(values) == 1:
                sys.exit()
            elif 'cont' in values:
                return
            else:
                print("Unknown suffix value --rdbg-show=" + str(values) + " => " + str(values[-1]))
                sys.exit()

    class OptActionSRCMAP(ActionExt):
        """Set the mapping of source file paths to executable file paths.
The input could be provided by one of the following syntaxes.

A single option value is supported in particular for the platforms *Solaris* and *OSX*,
but could be used on all.
The source and the exec paths are separated by the *os.pathsep*, which is splitted
by the *filesysobjects.apppaths.splitapppathx* [filesysobjects]_ function.

  --rdbg-srcmap := <one-entry-map>

    one-entry-map := (
         '"' <abs-source-filepath> '"' <srcmap-separator>  '"' <abs-exec-filepath> '"'
    )

As an alternate two separate option arguments for source path and executable path
on the other platforms such as Linux and NT / Windows.

  --rdbg-srcmap := <two-entries-map>

    two-entries-map := '"' <abs-source-filepath> '"'   '"' <abs-exec-filepath> '"'


The final values has to be transformed into a JSON compliant string of
lists-of-lists, where the inner represents a list with two elements.

  [["/src0", "/exec0"], ["/src1", "/exec1"], ["/src2", "/exec2"]]


This list is passed by *os.environ* to the module *pydev_file_utils.py*.

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionSRCMAP, self).__init__(option_strings, dest, nargs,
                                              **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            # preserves textual JSON list and inserts at the end
            if not namespace.rdbg_srcmap:
                namespace.rdbg_srcmap = _rdbgsrcmap

            if len(values) == 1:
                # single arg for a map - required for Solaris + OSX
                _smap = values[0].split(namespace.rdbg_srcmapsep)
                if len(_smap) != 2:
                    # s.th. went wrong...
                    raise PyRdbgError('parameter error --rdbg-srcmap=' + str(values))

            elif len(values) == 2:
                # two args for a map - e.g. Linux and NT
                _smap = values

            else:
                # s.th. went wrong...
                raise PyRdbgError('parameter error --rdbg-srcmap=' + str(values))

            if not namespace.rdbg_srcmap:
                namespace.rdbg_srcmap = [_smap]
            else:
                namespace.rdbg_srcmap.append(_smap)

    class OptActionSRCMAPSEP(ActionExt):
        """Changes separator for single-string source map values.
The value effects the following options '--rdbg-srcmap' only.

  --rdbg-srcmapsep <separator>

    default := ';'

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionSRCMAPSEP, self).__init__(option_strings, dest, nargs,
                                              **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            namespace.rdbg_srcmapsep = self.rdbg_srcmapsep = values[0]

    class OptActionHelp(ActionExt):
        """This help.

  --rdbg-help -rdbg-help

Context specific help is available for each option.

  --rdbg-... help
  --rdbg-... ?

For example:

  --rdbg-help ?

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionHelp, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            parser.print_help()
            sys.exit(0)

    class OptActionH(ActionExt):
        """\nUsage.

  --rdbg-h

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionH, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            print()
            parser.print_usage()
            sys.exit(0)

    class OptActionVERB(ActionExt):
        """Set the verbose display of RDBG states itself

  --rdbg-verb[=#level]

    level: "integer value of the verbose level"

Each repetition without a value increments the level by 1.

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionVERB, self).__init__(
                option_strings, dest, nargs, **kwargs)

            self.reset_val = False

        def call(self, parser, namespace, values, option_string=None):
            self.remove_option(parser, namespace, values, option_string)

            if values and values.isdigit():
                # reset early fetch for the case of '=' based value separators 
                rdbg.my_verb_self = int(values)
                self.reset_val = True

            elif self.reset_val:
                rdbg.my_verb_self += 1

            # from early fetch
            setattr(namespace, self.dest, rdbg.my_verb_self)

    class OptActionVersion(ActionExt):
        """\nCurrent version - terse.

  --rdbg-version

"""

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionVersion, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            print(str(__version__))
            sys.exit()

    class OptActionVersionDetails(ActionExt):
        """\nCurrent version - detailed.

  --rdbg-Version

        """

        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(OptActionVersionDetails, self).__init__(
                option_strings, dest, nargs, **kwargs)

        def call(self, parser, namespace, values, option_string=None):
            # if not values:
            #     namespace.optargs['Version'] = 0
            #     return
            # namespace.optargs['Version'] = values

            _width = 18
            _fstr = "%-" + str(_width) + "s: %s"
            print(
                "\n" + 
                _fstr % ("app", str(appname)) + 
                "\n" + 
                _fstr % ("id", str(__uuid__)) + 
                "\n" + 
                _fstr % ("pyrdbg", str(rdbg.__version__)) + 
                "\n" + 
                _fstr % ("rdbg.checkrdbg", str(__version__)) + 
                "\n" + 
                _fstr % ("Project", str('https://pyrdbg.sourceforge.net/')) + 
                "\n" + 
                _fstr % ("SCM", str('https://github.com/ArnoCan/pyrdbg/')) + 
                "\n" + 
                _fstr % ("package", str('https://pypi.python.org/pypi/pyrdbg/')) + 
                "\n" + 
                _fstr % ("docs", str(' https://pyrdbg.sourceforge.io/')) + 
                "\n" + 
                _fstr % ("author", str(__author__)) + 
                "\n" + 
                _fstr % ("author-www", str('https://arnocan.wordpress.com')) + 
                "\n" + 
                _fstr % ("copyright", str(__copyright__)) + 
                "\n" + 
                _fstr % ("license", str(__license__)) + 
                "\n"
            )
            sys.exit()

    class OptParser(object):
        """Option parser.
        """

        def __init__(self, argv=None, *args, **kargs):
            """
            Args:
                **argv**:
                    call options

            Returns:
                (opts, unknown, args)

            Raises:
                pass-through
            """
            # prep argv
            if argv and type(argv) in ISSTR:
                self.argv = argv.split()
            elif argv and type(argv) in (list,):
                self.argv = argv[:]
            elif argv is None:
                self.argv = sys.argv[:]
            else:
                self.argv = argv

            self.optargs = {}  # : resulting options and arguments'--
            self.opts_unknown = {}  # : resulting options not processed by actions'--

        def get_options(self):

            class FormatterWithCR(argparse.HelpFormatter):

                def _split_lines(self, text, width):
                    if text.startswith('@R:'):
                        return text[3:].splitlines()
                    return argparse.HelpFormatter._split_lines(self, text, width)

            if '--rdbg-h' in argv:
                self.parser = argparse.ArgumentParser(
                        add_help=False,
                        description="""
Injected help for the 'pyrdbg' debug interface.
""",
                        formatter_class=FormatterWithCR,
                    )
            else:
                self.parser = argparse.ArgumentParser(
                        add_help=False,
                        usage='%(prog)s [--rdbg [rdbg-options]] ...',
                        description="""
Injected help for the 'pyrdbg' debug interface.
""",
                        formatter_class=FormatterWithCR,
                    )

            group_options = self.parser.add_argument_group(
                'Remote Debug Server:',
                ''
            )
            group_options.add_argument(
                '--rdbg',
                '-rdbg',
                nargs='?',
                action=OptActionRDBGS,
                help="@R:" + OptActionRDBGS.__doc__,
            )
            group_options.add_argument(
                '--rdbg-forward',
                '-rdbg-forward',
                nargs='?',
                action=OptActionFWD,
                default=None,
                help="@R:" + OptActionFWD.__doc__
            )
            group_options.add_argument(
                '--rdbg-label',
                '-rdbg-label',
                nargs=1,
                action=OptActionLabel,
                help="@R:" + OptActionLabel.__doc__
            )
            group_options.add_argument(
                '--rdbg-path',
                '-rdbg-path',
                nargs='?',
                action=OptActionPATH,
                const=False,
                default=_rdbgpath_default,
                help="@R:" + OptActionPATH.__doc__
            )
            group_options.add_argument(
                '--rdbg-root',
                '-rdbg-root',
                nargs=1,
                action=OptActionROOT,
                const=_rdbgroot_default,
                help="@R:" + OptActionROOT.__doc__
            )
            group_options.add_argument(
                '--rdbg-srcmap',
                '-rdbg-srcmap',
                # nargs="2", # changed due to the behaviour of Solairs and OSX
                nargs="*",
                default=[],
                # type=str,
                action=OptActionSRCMAP,
                const=_rdbgsrcmap_default,
                help="@R:" + OptActionSRCMAP.__doc__
            )
            group_options.add_argument(
                '--rdbg-srcmapsep',
                '-rdbg-srcmapsep',
                # nargs="2", # changed due to the behaviour of Solairs and OSX
                nargs=1,
                default=';',
                type=str,
                action=OptActionSRCMAPSEP,
                const=_rdbgsrcmapsep_default,
                help="@R:" + OptActionSRCMAPSEP.__doc__
            )
            
            group_options.add_argument(
                '--rdbg-sub',
                '-rdbg-sub',
                nargs='?',
                action=OptActionSUB,
                const=_rdbgsub_default,
                help="@R:" + OptActionSUB.__doc__
            )
            group_options.add_argument(
                '--rdbg-config',
                '-rdbg-config',
                nargs='?',
                action=OptActionConfig,
                default=None,
                help="@R:" + OptActionConfig.__doc__
            )
            group_options.add_argument(
                '--rdbg-env',
                '-rdbg-env',
                nargs='?',
                action=OptActionEnv,
                default=None,
                help="@R:" + OptActionEnv.__doc__
            )

            group_options = self.parser.add_argument_group(
                'RDBG test:',
                ''
            )
            group_options.add_argument(
                '--rdbg-keep-argv',
                '-rdbg-keep-argv',
                nargs='?',
                action=OptActionKeepArgv,
                default=None,
                help="@R:" + OptActionKeepArgv.__doc__
            )
            group_options.add_argument(
                '--rdbg-options',
                '-rdbg-options',
                nargs='*',
                action=OptActionPrintOptions,
                default=[],
                help="@R:" + OptActionPrintOptions.__doc__
            )
            group_options.add_argument(
                '--rdbg-self',
                '-rdbg-self',
                nargs='?',
                action=OptActionDBG,
                default=None,
                const=None,
                help="@R:" + OptActionDBG.__doc__
            )
            group_options.add_argument(
                '--rdbg-testflags',
                '-rdbg-testflags',
                nargs=1,
                action=OptActionTFLAG,
                type=str,
                # default=[],
                const=[],
                help="@R:" + OptActionTFLAG.__doc__
            )

            group_options = self.parser.add_argument_group(
                'Support Options:',
                ''
            )
            group_options.add_argument(
                '--rdbg-get-stub',
                '-rdbg-get-stub',
                nargs='*',
                default=[],
                action=OptActionGetStub,
                help="@R:" + OptActionGetStub.__doc__
            )
            group_options.add_argument(
                '--rdbg-get-stub-mod',
                '-rdbg-get-stub-mod',
                nargs='*',
                default=[],
                action=OptActionGetStubMod,
                help="@R:" + OptActionGetStubMod.__doc__
            )
            group_options.add_argument(
                '--rdbg-show',
                '-rdbg-show',
                nargs='*',
                default=[],
                action=OptActionSHOW,
                help="@R:" + OptActionSHOW.__doc__
            )
            group_options.add_argument(
                '--rdbg-version',
                '-rdbg-version',
                nargs=0,
                action=OptActionVersion,
                default=None,
                help="@R:" + OptActionVersion.__doc__
            )
            group_options.add_argument(
                '--rdbg-Version',
                '-rdbg-Version',
                nargs=0,
                action=OptActionVersionDetails,
                default=None,
                help="@R:" + OptActionVersionDetails.__doc__
            )
            group_options.add_argument(
                '--rdbg-h',
                '-rdbg-h',
                nargs='?',
                action=OptActionH,
                default=None,
                help="@R:" + OptActionH.__doc__
            )
            group_options.add_argument(
                '--rdbg-help',
                '-rdbg-help',
                nargs='?',
                action=OptActionHelp,
                default=None,
                help="@R:" + OptActionHelp.__doc__
            )
            group_options.add_argument(
                '-rdbg-... help',
                '--rdbg-... ?',
                action=None,
                metavar='',
                help="@R:" + """Context specific help for each option.

  --rdbg-... help
  --rdbg-... ?

For example:

  --rdbg-help ?

"""
            )

            group_see = self.parser.add_argument_group(# @UnusedVariable
                'SEE ALSO:',
                """
                https://pypi.python.org/pypi/pyrdbg/,
                https://pyrdbg.sourceforge.io/'
                """
            )

            group_copyright = self.parser.add_argument_group(# @UnusedVariable
                'COPYRIGHT:',
                'Arno-Can Uestuensoez (C)2015-2018 @Ingenieurbuero Arno-Can Uestuensoez',
            )

            #
            # opts contains a tuple of the populated namespace with knonw options and
            # the remaining options not processed by actions objects
            res = self.parser.parse_known_args(argv, namespace=optspace)

            self.opts = optspace
            self.opts_unknown = res[1]
            if rdbg.my_dbg_self > 1 and not optspace.rdbg_options:
                sys.stderr.write(
                    "RDBG:%s:%d:CLI:options: %s\n" % (
                        __name__,
                        getcaller_linenumber(),
                        self.opts.__str__(indent=3))
                      )

            if optspace.rdbg_options:
                sys.stdout.write("RDBG:CLI:options:\n")
                if optspace.rdbg_options[0] == 'str':
                    for k, v, in optspace.get_items():
                        sys.stdout.write("  %-20s: %s\n" % (k, v,))

                elif optspace.rdbg_options[0] == 'repr':
                    sys.stdout.write("%s\n" % (repr(optspace.get_JSON_bin())))
                
                elif self.opts.rdbg_options[0] == 'json':
                    import jsondata.jsondata  # on demand only
                    sys.stdout.write(
                        "%s\n" % (str(jsondata.jsondata.JSONData(optspace.get_JSON_bin()))))
                
                else:
                    raise PyRdbgError("Output format not supported for options print:" + str(self.opts.out_format))

                if self.opts.rdbg_options[1] == 'break':
                    sys.exit(0)

            elif rdbg.my_dbg_self > 3:
                sys.stderr.write("RDBG:%s:%d:CLI:options:\n" % (
                    __name__,
                    getcaller_linenumber(),
                    ))
                import jsondata.jsondata  # @Reimport  # on demand only
                print(str(jsondata.jsondata.JSONData(optspace.get_JSON_bin())))

            if optspace.rdbg_show:
                OptActionSHOW.show_opt(optspace, optspace.rdbg_show)
                sys.exit(0)

            return res

        def print_usage(self, *args):
            self.parser.print_usage(*args)

        def print_help(self, *args):
            self.parser.print_help(*args)

    return OptParser(argv, **kargs).get_options()


def checkandremove_rdbgoptions(argv=None, **kargs):
    """Checks input options from sys.argv and returns the tuple of
    resulting debug parameters. The options are removed from sys.argv
    by default.

    The options are kept as provided, or when missing filled with default values.
    The final resolution of the values like file system globs is performed within
    the PyDevRDC class.

    The following options are checked and removed from sys.argv, thus by default
    has to be added by the caller application again when required for following
    nested calls, for details refer to the manuals
    `CLI Options <rdbg_cli_options.html>`_ : ::

       --rdbg [[host]:[port]]
       --rdbg-Version
       --rdbg-config <config-file>
       --rdbg-env
       --rdbg-forward [<depth>]
       --rdbg-keep-argv
       --rdbg-label <label>
       --rdbg-path (<literal>|<pathforscan>|<FQDN>|<path-glob>)
       --rdbg-root (<rootforscan>|<FQDN>|<path-glob>)
       --rdbg-show [('all'|'load'|'opt'|'rdbg'|'rt')]
       --rdbg-srcmap <list-of-path-mapping>
       --rdbg-sub (<literal>|<path-glob>)
       --rdbg-version

    Context Help: ::

       --rdbg-h - short usage
       --rdbg-help - detailed help
       --<any-option> ? - help on the <any-option> only

    Test options and flags: ::

       --pderd-test-suppress-init
       --rdbg-get-stub
       --rdbg-get-stub-mod
       --rdbg-options ('json' | 'repr' | 'str')
       --rdbg-self[=#level]
       --rdbg-testflags[('ignorePydevd'|'ignorePydevdSysPath'|'ignorePydevdCallParam')]

    Environment variables:
        The provided environment variables are read out, when no related
        option is available.

    Args:
        **argv**:
            Alternative argv.

            *default:=sys.argv*

        kargs:
            **debug**:
                Same as '--debug', developer output.

            **fpname**:
                The file path name of the main file for the current
                process,

                *default:=callname*

            **label**:
                The label identifying the current process in the
                debug forwarding into subprocesses.

                *default := callname*

            **namespace**:
                Namespace with extracted options. Sets options as
                super position of the values from hard-coded data,
                configuration file, environment variables, and finally
                of the command line parameters.

            **rdbg**:
                Same as '--rdbg'.

            **rdbgforward**:
                Same as '--rdbg-forward'.

            **rdbgenv**:
                Same as '--rdbg-env'.

            **rdbgkeepargv**:
                Same as '--rdbg-keep-argv'.

            **rdbglabel**:
                Same as '--rdbg-label'.

            **rdbgpath**:
                Same as '--rdbg-path'.

            **rdbgroot**:
                Same as '--rdbg-root'.

            **rdbgsrcmap**:
                Same as '--rdbg-srcmap'.

            **rdbgsub**:
                Same as '--rdbg-sub'.

            **verbose**:
                Same as '--verbose', user output.

    Global Attributes:
        Sets the global attributes: ::

            _rdbgthis        # debug this process
            _rdbgsrv         # the RDS
            _rdbgfwd         # forward the debug-enabler to subprocess
            _rdbgpath        # the path of the debug stub module
            _rdbgroot        # the root path for the debug stub module
            _rdbgsub         # the subdirectory pattern of the debug stub module
            _rdbgsrcmap      # map list of executables/libraries and associated sources
            _rdbglabel       # label of this process for the forwarding chain
            _rdbgtestflags   # test-mode for RDBG itself

    Returns:
        When successful returns tuple of the following debug parameters. The
        options are removed from 'sys.argv'. ::

           return_value := {
                'rdbg':           rdbgsrv,
                'rdbgfwd':        rdbgfwd,
                'rdbglabel':      rdbglabel,
                'rdbgpath':       rdbgpath,
                'rdbgroot':       rdbgroot,
                'rdbgsrcmap':     rdbgsrcmap,
                'rdbgsub':        rdbgsub,
                'rdbgtestflags':  rdbgtestflags,
                'rdbgthis':       rdbgthis,
                'pydevd':         pydevd,
            }

            pydevd': pydevd,
            rdbgthis = (True|False)        # Debugging of current process/thread is enabled.
            rdbg = [host][:port]           # Connection parameter of peer RemoteDebugServer
            rdbgfwd = <partialprocessed>   # Forward debug parameters to subprocesses
               partialprocessed: (         # One of:
                  hop = hopcnt - 1         #   level of subrocesses
                  | 'all'                  #   all - not yet available
                  | label                  #   the subprocess with the assigned label
               )

            rdbgroot = <rootforscan>       # Base for search, when None the defaults apply.
            rdbgsub = <eclipse-subpath>    # Subpath within Eclipse.
            rdbgpath = <pydevd-path>       # Value of option rdbgpath and/or resolved path
                                           # from rdbgroot and rdbgsub.
            rdbgsrcmap = <list-of-maps>    # List of path mappings.
            rdbglabel = <forwarding-label> # User defined label.
            pydevd = <current-module>      # Reference to loaded module, else None.

    Raises:
        *RdbgNetworkError*

        *RdbgCheckOptionsError*

    """
    global _rdbgthis
    global _rdbgsrv
    global _rdbgfwd
    global _rdbgpath
    global _rdbgroot
    global _rdbgsub
    global _rdbgsrcmap
    global _rdbglabel
    global _rdbgtestflags

    _rdbgenv = kargs.get('rdbgenv', False)
    
    if kargs.get('debug'):
        rdbg.my_dbg_self = kargs.get('debug')

    _rdbg = False
    _lbl = None
    _fpname = None
    _pydevd = None
    _forceload = False
    _enableautoload = False
    _enableautopath = None
    
    for k, v in kargs.items():
        if k == 'label':
            _lbl = v

        elif k == 'rdbglabel':
            _rdbglabel = v

        elif k == 'fpname':
            _fpname = v

        elif k == 'enableautoload':
            # enables headless auto load feature
            _enableautoload = True
            _rdbg = True
            _rdbgthis = True

        elif k == 'enableautopath':
            _enableautopath = v

        elif k == 'forceload':
            _forceload = True

        elif k == 'rdbg':
            # enables standard call
            _rdbg = True
            _rdbgthis = True
            _rdbgsrv = v

        elif k == 'rdbgsrv':
            _rdbg = True
            _rdbgsrv = v

        elif k == 'rdbgforward':
            _rdbg = True
            _rdbgfwd = v

        elif k == 'rdbgroot':
            _rdbgroot = v

        elif k == 'rdbgsrcmap':
            _rdbgsrcmap = v

        elif k == 'rdbgsub':
            _rdbgsub = v

        else:
            raise Exception("Unknown option:" + str(k))

    if _enableautopath:

        # ensure auto load and auto params 
        if not _enableautoload:
            raise PyRdbgError("'enableautopath' is only valid with 'enableautoload'")

        if _enableautopath == None:
            _enableautopath = os.curdir
        
        if not os.path.exists(_enableautopath):
            raise PyRdbgError("missing enableautopath = " + str(_enableautopath))

    # prep argv for extraction of rdbg options
    if argv and type(argv) in ISSTR:
        _ty = argv.split()
    elif argv == None:
        _ty = sys.argv
    else:
        _ty = argv

    # label of current process
    if not _lbl:
        _lbl = os.path.basename(_ty[0])
    if not _fpname:
        _fpname = os.path.abspath(_ty[0])

    #
    # read own options
    #
    optspace = kargs.get('namespace')
    if optspace == None:
        optspace = RdbgSpace(
            _ty,
            enableautoload=_enableautoload,
            enableautopath=_enableautopath 
            )
        
    _clibuf = fetch_rdbg_opts(_ty, optspace=optspace, **kargs)
    if _clibuf[0].rdbg or _clibuf[0].rdbg_forward or _forceload:
        # standard activation
        _rdbg = True

#     elif not _rdbg:
#         # had been enabled for activation - auto load
#         _rdbg = False
#         if _clibuf[0].rdbg_root:
#             raise RdbgCheckOptionsError(
#                 "Missing activation option '--rdbg', provided:'--rdbg-root'")
# 
#         if _clibuf[0].rdbg_sub:
#             raise RdbgCheckOptionsError(
#                 "Missing activation option '--rdbg', provided:'--rdbg-sub'")

    if _rdbg:
        _rdbgsrv = _clibuf[0].rdbg

        # common network configuration
        try:
            hn0 = socket.gethostname()
            if rdbg.my_dbg_self > 1:
                sys.stderr.write(
                    "RDBG:" + __name__ + ":socket.gethostname()   = " + str(hn0) + "\n"
                )
        except Exception as e:
            raise RdbgNetworkError(str(e) + "\nERROR:NETWORK:Cannot get host name by socket.gethostname()")

        # common network configuration
        try:
            hn = socket.gethostbyaddr(hn0)
            if rdbg.my_dbg_self > 1:
                sys.stderr.write(
                    "RDBG:" + __name__ + ":_socket.gethostbyaddr() = " + str(hn) + "\n"
                )
        except Exception as e:
            raise RdbgNetworkError(
                str(e) 
                +"\nERROR:NETWORK:Cannot get host name by socket.gethostbyaddr(" + str(hn0) + ")"
                )

        if _clibuf[0].rdbg is None:
            _rdbgthis = True

        elif _clibuf[0].rdbg.startswith(
                'localhost') or _clibuf[0].rdbg.startswith('127.0.0.1'):
            _rdbgthis = True

        elif _clibuf[0].rdbg == hn[0] or _clibuf[0].rdbg in hn[1] or _clibuf[0].rdbg in hn[2]:
            _rdbgthis = True

        if _clibuf[0].rdbg_forward != None:
            _rdbgfwd = _clibuf[0].rdbg_forward

        if _clibuf[0].rdbg_label != None:
            _rdbglabel = _clibuf[0].rdbg_label

        if _clibuf[0].rdbg_path != None:
            _rdbgpath = _clibuf[0].rdbg_path

        if _clibuf[0].rdbg_root != None:
            _rdbgroot = _clibuf[0].rdbg_root

        if _clibuf[0].rdbg_sub != None:
            _rdbgsub = _clibuf[0].rdbg_sub

        if _clibuf[0].rdbg_srcmap != None:
            _rdbgsrcmap = _clibuf[0].rdbg_srcmap

        if _clibuf[0].rdbg_self != None:
            rdbg.my_dbg_self = _clibuf[0].rdbg_self

        if _clibuf[0].rdbg_testflags != None:
            _rdbgtestflags = _clibuf[0].rdbg_testflags

    if rdbg.my_dbg_self > 1:
        _lnum = getcaller_linenumber()
        sys.stderr.write(
            "RDBG:%s:%d:_rdbgenv    = %s\n"
            "RDBG:%s:%d:_rdbgfwd    = %s\n"
            "RDBG:%s:%d:_rdbglabel  = %s\n"
            "RDBG:%s:%d:_rdbgpath   = %s\n"
            "RDBG:%s:%d:_rdbgroot   = %s\n"
            "RDBG:%s:%d:_rdbgsrcmap = %s\n"
            "RDBG:%s:%d:_rdbgsrv    = %s\n"
            "RDBG:%s:%d:_rdbgsub    = %s\n"
            "RDBG:%s:%d:_rdbgtestflags = %s\n"
            "RDBG:%s:%d:_rdbgthis   = %s\n" % (
                __name__, _lnum, str(_rdbgenv),
                __name__, _lnum, str(_rdbgfwd),
                __name__, _lnum, str(_rdbglabel),
                __name__, _lnum, str(_rdbgpath),
                __name__, _lnum, str(_rdbgroot),
                __name__, _lnum, str(_rdbgsrcmap),
                __name__, _lnum, str(_rdbgsrv),
                __name__, _lnum, str(_rdbgsub),
                __name__, _lnum, str(_rdbgtestflags),
                __name__, _lnum, str(_rdbgenv),
            )
        )

        try:
            _pydevd = sys.modules['pydevd']
            if 'ignorePydevd' in _rdbgtestflags:
                sys.stderr.write(
                    "RDBG:" + __name__ + ":" + str(getcaller_linenumber()) 
                    +":ignore by flag:pydevd pre-loaded = " + str(_pydevd) + "\n"
                    )
            else:
                sys.stderr.write(
                    "RDBG:" + __name__ + ":" + str(getcaller_linenumber())
                    +":pydevd pre-loaded = " + str(_pydevd) + "\n"
                    )
        except:
            sys.stderr.write(
                "RDBG:" + __name__ + ":" + str(getcaller_linenumber())
                +":pydevd pre-loaded = False\n"
                )
            pass

    _ty = _clibuf[1]

    # forwards debugging state to nested subprocesses
    if _rdbgfwd and not _rdbgthis:
        if type(_rdbgfwd) is str and _rdbgfwd.lower() == 'all':  # all
            _rdbgthis = True

        elif type(_rdbgfwd) == int and _rdbgfwd:  # hopcount
            if _rdbgfwd >= 0:  # debug the first:  [0:hopcnt]
                _rdbgfwd -= 1

                if _rdbgfwd >= 0:
                    _rdbgthis = True

                else:
                    _rdbgthis = False

            else:  # debug from hopcnt on: [hopcnt:]
                _rdbgfwd += 1

                if _rdbgfwd < 0:
                    _rdbgthis = False

                else:
                    _rdbgthis = True

        elif type(_fpname) is str and os.path.exists(_fpname):  # fpname

            if os.path.abspath(_fpname) == os.path.abspath(_ty[0]):
                _rdbgthis = True

            else:
                _rdbgthis = False

        elif type(_rdbgfwd) is str:  # label
            if _rdbgfwd == _lbl:
                _rdbgthis = True

    #
    # set now missing as defaults
    #
    if _rdbgthis and not _rdbgsrv:
        _rdbgsrv = _rdbg_default

    if _rdbgenv:  # if set, force when present
        if os.environ.get('RDBGPATH'):
            _rdbgpath = os.environ['RDBGPATH']
        if os.environ.get('RDBGROOT'):
            _rdbgroot = os.environ['RDBGROOT']
        if os.environ.get('RDBGSUB'):
            _rdbgsub = os.environ['RDBGSUB']

    if rdbg.my_dbg_self > 2:
        sys.stderr.write(
            "RDBG:" + __name__ + ":" + str(getcaller_linenumber()) + 
            ":_rdbgpath = " + str(_rdbgpath) + "\n"
            )

    if _rdbgpath == False:
        # 1. check for a bundled installation with PyDev in Eclipse
        _rdbgpath = scan_for_pydevd(_rdbgroot, altpat=_rdbgsub)

    elif not _rdbgpath:
        if 'ignorePydevdSysPath' in _rdbgtestflags:
            # mainly for the test of RDBG itself
            _rdbgpath = scan_for_pydevd(_rdbgroot, altpat=_rdbgsub)

        else:
            # 0. check for an *pydevd* standalone installation
            try:
                _rdbgpath = filesysobjects.pathtools.findrelpath_in_searchpath('pydevd.py')
                _rdbgpath = os.path.dirname(_rdbgpath)
    
            except Exception as e:
    
                # 1. check for a bundled installation with PyDev in Eclipse
                _rdbgpath = scan_for_pydevd(_rdbgroot, altpat=_rdbgsub)

    else:
        _r = _rdbgpath
        _rdbgpath = expandpath(*_rdbgpath)
        if not _rdbgpath:
            raise RdbgScanPydevdError("File path does not exist:" + str(_r))
#        _rdbgpath = _rdbgpath[0]

    ret = {
        'rdbgthis': _rdbgthis,
        'rdbg': _rdbgsrv,
        'rdbgfwd': _rdbgfwd,
        'rdbgroot': _rdbgroot,
        'rdbgsub': _rdbgsub,
        'rdbgpath': _rdbgpath,
        'rdbgsrcmap': _rdbgsrcmap,
        'rdbglabel': _rdbglabel,
        'rdbgtestflags': _rdbgtestflags,
        'pydevd': _pydevd,
    }

    if rdbg.my_dbg_self > 2:
        sys.stderr.write(
            "RDBG:" + __name__ + ":" + str(getcaller_linenumber()) + 
            ":rdbgoptions = " + str(ret) + "\n"
            )

    if _rdbgsrcmap:  # _rdbgsrcmap
        _srcmap = '['
        for x in _rdbgsrcmap:
            _srcmap += '["' + str(x[0]) + '","' + str(x[1]) + '"],'
        _rdbgsrcmap = _srcmap[:-1] + ']'
        os.environ['PATHS_FROM_ECLIPSE_TO_PYTHON'] = _rdbgsrcmap

        if rdbg.my_dbg_self:
            sys.stderr.write(
                "RDBG:%s:%d:setting source map:%s\n" % (
                __name__, getcaller_linenumber(), _rdbgsrcmap))

    #
    # workout display commands
    #
    if _clibuf[0].rdbg_get_stub_mod:
        sys.stdout.write(ret['rdbgpath'])
        sys.exit(0)
    
    elif _clibuf[0].rdbg_get_stub:
        sys.stdout.write(os.path.dirname(ret['rdbgpath']))
        sys.exit(0)

    return ret


def checkrdbg(argv=None):
    """Checks presence of '--rdbg' option.

    Args:
        **argv**:
            Alternative argv.

            *default:=sys.argv*

    Returns:
        *True*:  RDBG setup to active
        
        *False*: else

    Raises:
        *None*

    """
    if argv == None:  # not set at all
        return '--rdbg' in sys.argv
    elif '--rdbg' in argv:  # simple '--rdbg', or without '='
        return True

    # '--rdbg' with '='
    return True in map(lambda x: x.startswith('--rdbg='), argv)


def getdefaultproxy():
    """Gets the shared default remote debug controller proxy.

    Args:
        *None*

    Returns:
        Default proxy instance, the shared static instance of
        *rdbg.pydevrdc.PyDevRDC*.

        The value is *None* when not yet initialized.

    Raises:
        *None*

    """
    return rdbg.shared.PYDEVD


def getdefaults(res=None):
    """Returns the current defaults.

    Args:
        **res**:

            Result, the type of 'res' defines the return type.
            The default when 'None' is tuple.

    Returns:
        Returns the collection of default. The following options are
        available: ::

           res == None: (default)
                # result is a new created tuple:
                res := (,)

           res in (tuple, list,):
                # see default, container type as provided.
                res == dict:
                res := { }

           res is rdbg.Namespace
             # Similar to argparse.Namespace, foreseen to be used also
             # as predefined default input for 'rdbg.checkrdbg'.

    """
    return {
        'rdbg': _rdbg_default,
        'rdbgfwd': _rdbgfwd_default,
        'rdbglabel': _rdbglabel_default,
        'rdbgpath': _rdbgpath_default,
        'rdbgroot': _rdbgroot_default,
        'rdbgsrcmap': _rdbgsrcmap_default,
        'rdbgsrcmapsep': _rdbgsrcmapsep_default,
        'rdbgsub': _rdbgsub_default,
        'rdbgtestflags': [],
        'rdbgthis': False,
        'pydevd': pydevd_module,
    }


def getnamespace():
    """Returns a reference to the current loaded namespace.
    The namespace object contains the complete set of rdbg-call
    options with interfaces to read out.

    Args:
        *None*

    Returns:
        The reference to the contained name space containing
        the complete set of current resulting parameters.

    Raises:
        pass-through

    """
    return


def setdefaultproxy(dproxy):
    """Sets the shared default remote debug controller proxy.

    Args:
        **dproxy**:
            The new default proxy. requires an instance of type
            *rdbg.pydevrdc.PyDevRDC*.

    Returns:
        The old proxy. The value is 'None', when this was not yet
        initialized.

    Raises:
        *None*

    """
    return rdbg.pydevrdc.setdefaultproxy(dproxy)


def isactive(rdbgobj=None):
    """Checks current status of rdbg.

    The remote debugging is managed by PyDev via *pydevd_io.IORedirector*
    thus programs called from the command line do not detect a 'tty'.
    The current status of rdbg could be used as additional criteria for
    the debugging of console related functions.

    Args:
        *None*

    Returns:
        *True*:   Running debug session initiated by 'rdgb'.

        *False*:  No 'rdbg' initiated session.

    Raises:
        pass-through

    """
    if rdbgobj == None:
        if getdefaultproxy():
            return getdefaultproxy().isactive()
    return False


def isattached(rdbgobj=None):
    """Checks whether started from within PyDev or as standalone program attached to/by rdbg.
    Check whether *pydevd* is already preloaded, if not decides to be attached.

    Args:
        *None*

    Returns:
        *True*:   Running debug session initiated by 'rdgb'.

        *False*:  No 'rdbg' initiated session.

    Raises:
        pass-through

    """
    if rdbgobj == None:
        if getdefaultproxy():
            return getdefaultproxy().isattached()
    return False


def ispydevcalled(rdbgobj=None):
    """Checks whether started from within PyDev.
    Check whether *pydevd* is already preloaded, if not decides to be attached.

    Args:
        **rdbgobj**:
            The proxy instance of PyDevRDC to be checked,
            else *sys.modules['pydevd']*.

    Returns:
        *True*:   Started by *PyDev* e.g. from within *Eclipse*.

        *False*:  Session created from external, e.g. via PyDev-Debug-Server.

    Raises:
        pass-through
    """
    if rdbgobj:
        return rdbgobj.ispydevcalled()
    else:
        return callfromPyDevDbg


def setdefaults(res=None):
    """Sets defaults.

    """
    global _rdbgthis
    global _rdbgsrv
    global _rdbgfwd
    global _rdbgroot
    global _rdbgsub
    global _rdbgsrcmap
    global _rdbglabel

    _rdbgthis = False
    _rdbgsrv = _rdbg_default
    _rdbgfwd = _rdbgfwd_default
    _rdbgroot = _rdbgroot_default
    _rdbgsrcmap = _rdbgsrcmap_default
    _rdbgsrcmapsep = _rdbgsrcmapsep_default
    _rdbgsub = _rdbgsub_default

    return True



_clrargs = {
    'abs': True,
    'non-existent': True,
    'split': True,
    'redundant': True,
    'shrink': True,
    'normpath': True
}
"""
Args for 'filesysobjects.clearpath'
"""


def scan_for_pydevd(rootforscan=None, **kargs):
    """Scans file system directory tree for the debug stub, by default
    'pydevd' for the package *PyDev* as a plugin for *Eclipse*.

    See PyDev manual for path reference, the default pattern for the
    'marketplace' based installation is: ::

       eclipse/plugins/org.python.pydev_x.x.x/pysrc/pydevd.py

    In case of drop-in installation ::

       eclipse/dropins/<pydev-dropin-name>/plugins/org.python.pydev_x.x.x/pysrc/pydevd.py

    The matching versions could be varied by glob-expressions.
    The provided parameters match as follows: ::

       rdbgroot := /path/to/eclipse
       rdbgsub  := org.python.pydev_x.x.x/pysrc/pydevd.py

    The glue-hook depends on the type of installation and is determined
    dynamically: ::

       dropin-install       := dropins/<pydev-dropin-name>/plugins
       market-place-install :=  plugins

    Basically any path could be used, in particular a rdbgsub directory
    containing a subset for 'pydevd.py' on a remote machine as stated
    by the PyDev project for remote debugging of server processes. The
    filesystem resolution is performed on a local filesystem only, but
    could be performed by command line start of the headless process
    on the remote machine too.

    The path is the containment path of the file *pydevd.py* and has to be
    included in the *sys.path* variable for activation.

    The *scan_for_pydevd* method performs a search and returns the
    absolute path in case of a match.

    The search for the root directory into the Eclipse package installation
    is performed in the following order and priority:

    1. **parameters**

        Consume call/command line parameters, ENV, and hard/coded.

        Parameter mix by environment variables as present:
        For the actual main control of environment variables
        refer to rdbg.checkrdbg.checkandremove_rdbgoptions. ::

           (rdbgroot or RDBGROOT) + ( rdbgsub or RDBGSUB )

        Contains the value for the option '--rdbg-root',
        either literal or as a 'glob' pattern.
        The following priorities are applied:

        1. CLI call option
        2. API call option
        3. RDBGROOT/RDBGSUB + missing from
        4. Code defaults

    2. **sys.path**

       Each separate path is tried with the sub-path pattern,
       first match wins.

    3. **search eclipse installation**

       Each separate path is tried with the sub-path pattern,
       calls: ::

          scan_eclipse_for_pydevd

    The first match of containing directory for 'pydevd.py'
    is returned, thus ambiguity in case of multiple occurrences
    has to be avoided.

    The pattern resolution into the PyDev dir is performed
    by the steps: ::

       0. <eclipse-root>/plugins/<rdbgsub>
       1. <eclipse-root>/dropins/<rdbgsub>
       2. <eclipse-root>/dropins/*/plugins/<rdbgsub>

    For additional details refer to the section 
    'Client Stub Lookup for "pydevd"' in the manual.

    Args:
        **rootforscan**:
            Start directory for tree-scan:

            * a single path string
            * search path notation with multiple entries
            * a *list* of paths

            Each path points to an Eclipse installation
            directory.

    kargs:
        **altpat**:
            Alternative pattern, varies 'eclipse_glob',
            and 'pydevd_glob'. ::

               altpat := (<literal> | <glob>)

        **strict**:
            Provided parameter has to match, else error.

            Default is to try all, when supplied params do not
            match default values are checked.

        **version**:
            Provide version to be requested. It is recommended
            to combine this with 'strict'.

            The parameters 'altpat' and 'version' are EXOR. ::

               version=(a,b,c)
               
               a := [0-9]+
               b := [0-9]+
               c := [0-9]+

    Returns:
        The location of pydevd.py

    Raises:
        *AttributeError*

    """
    pydevd_glob = _rdbgsub
    _itfp = False
    _itfs = False
    _itfcp = False

    if _rdbgtestflags:  # CLI interaction has priority
        for tf in _rdbgtestflags:
            if tf == 'ignorePydevd':
                _itfp = True
            elif tf == 'ignorePydevdSysPath':
                _itfs = True
            elif tf == 'ignorePydevdCallParam':
                _itfcp = True

    _mlist = []  # list of matches

    _apat = kargs.get('altpat')
    _vers = kargs.get('version')
    _pg = pydevd_glob

    def _matchsub(_apat, _rfs):
        """
        Args:
            **_apat**:
                Alternate relative file path name pattern with wildcards.
                For provided *glob* and *re* see *expandpath*.
                The scans are: ::

                   0: any directory               : _rfs/_apat
                   1: as IDE-plugin sub directory : _rfs/plugins/_apat
                   2: as IDE-dropin sub directory : _rfs/dropins/_apat

            **_rfs**:
                Root directory for path match.

        Returns:
            *Success*:  file path

            *Failure*:  None

        Raises:
            pass-through

        """

        if rdbg.my_dbg_self > 2:
            sys.stderr.write(
                "RDBG:%s:%d:_matchsub:sub dir pattern:\n   root   = %s\n   sub    = %s\n" % (
                __name__, getcaller_linenumber(),
                str(_rfs),
                str(_apat),
                ))

        # almost literal
        p = expandpath(_rfs + os.sep + _apat, **{'wildcards': True})
        if not p:
            # plugins
            p = expandpath(_rfs + os.sep + "plugins" + os.path.sep + _apat, **{'wildcards': True})
            if not p:  # a match trial
                # dropins
                p = expandpath(_rfs + os.sep + "dropins" + os.path.sep + "*" + os.path.sep
                               +"plugins" + os.path.sep + _apat, **{'wildcards': True})
        if rdbg.my_dbg_self > 2:
            sys.stderr.write(
                "   result = %s\n" % (
                str(p),
                ))

        return p

    # 0. fetch call parameters which also reflect the resulting command line parameters
    if _apat and _vers:
        raise RdbgScanPydevdError("One only supported (version, altpat)=('"
                                      +str(_vers) + "', '" + str(_apat) + 
                                      "')")
    elif _vers:
        _apat = "org.python.pydev_" + \
            _vers[0] + "." + _vers[1] + "." + _vers[2] + "*/pysrc/pydevd.py"

    elif not _apat:
        if pydevd_glob:
            _apat = pydevd_glob

        else:
            _apat = _rdbgsub_default

    if not _apat:
        raise RdbgScanPydevdError("Cannot evaluate rdbgsub-pattern")

    _st = kargs.get('strict', False)

    if rootforscan == None:  # parameter not provided, so default
        _rfs = _rdbgroot

    elif type(rootforscan) != list:  # is a string
        _rfs = [rootforscan]  # for in-place parameter passing
        clearpath(_rfs, **_clrargs)
        _rfs = _rfs[0]

    else:  # is a list
        _rfs = rootforscan[:]  # for parameters passed in-place - keep them unchanged
        clearpath(_rfs, **_clrargs)
        if _rfs:
            _rfs = _rfs[0]
        else:
            raise PyRdbgError("Provided root-path does not exist:" + str(rootforscan))

    if not _rfs:
        _rfs = _rdbgroot_default
    if not _rfs:
        raise RdbgScanPydevdError("Cannot evaluate rdbgroot for scan:" + str(rootforscan))

    if rdbg.my_dbg_self > 2:
        sys.stderr.write(
            "RDBG:%s:%d:root for scan: %s\n" % (
            __name__, getcaller_linenumber(),
            str(_rfs),
            ))

    p = None

    #
    # 1. call parameter
    #
    if not _itfcp:

        if rdbg.my_dbg_self:
            sys.stderr.write(
                "RDBG:%s:%d:check:call parameter: %s\n" % (
                __name__, getcaller_linenumber(),
                str(_rfs),
                ))

        p = _matchsub(_apat, _rfs)  # expects one only...
        if p:
            return p
        if _rfs and _st:  # has to match immediately when STRICT is choosen
            return

    #
    # 2. environment RDBGROOT and/or RDBGSUB
    #
    if not p:
        try:
            if _rdbgenv:
                _r = os.environ['RDBGROOT']
                if not _r:
                    _r = _rfs
                else:
                    if rdbg.my_dbg_self > 2:
                        sys.stderr.write(
                            "RDBG:%s:%d:environ: RDBGROOT = %s\n" % (
                            __name__, getcaller_linenumber(),
                            str(_r),
                            ))
                    _r = [
                        _r,
                    ]

                _s = os.environ['RDBGSUB']
                if not _s:
                    _s = _apat
                else:
                    if rdbg.my_dbg_self > 2:
                        sys.stderr.write(
                            "RDBG:%s:%d:environ: RDBGSUB = %s\n" % (
                            __name__, getcaller_linenumber(),
                            str(_s),
                            ))

                sys.stderr.write(
                    "RDBG:%s:%d:check:environ:\n   %s\n   %s\n" % (
                    __name__, getcaller_linenumber(),
                    str(_r),
                    str(_s),
                    ))

                clearpath(_r, **_clrargs)
                p = _matchsub(_s, _r)
                if p:
                    return p
                elif _r and _s:  # has to match immediately when STRICT is choosen
                    return
            else:
                _r = _rfs
                _s = _apat

        except Exception:
            pass

    #
    # following are the preset and hard-coded defaults
    #

    #
    # 3. search 'pydevd.py' by sys.path
    #
    if not _itfs and not rootforscan:
        sys.stderr.write(
            "RDBG:%s:%d:check:sys.path for 'pydevd': %s\n" % (
            __name__, getcaller_linenumber(),
            str(_r),
            ))
        for pi in sys.path:
            p = _matchsub('pydevd.py', pi)
            if p:
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:sys.path for 'pydevd.py': sys.path[x] = %s\n" % (
                        __name__, getcaller_linenumber(),
                        str(p),
                        ))
                return p
            else:
                p = _matchsub('pydevd', pi)
                if p:
                    if rdbg.my_dbg_self > 2:
                        sys.stderr.write(
                            "RDBG:%s:%d:sys.path for 'pydevd': sys.path[x] = %s\n" % (
                            __name__, getcaller_linenumber(),
                            str(p),
                            ))
                    return p

    # 4. search pattern by sys.path
    if not _itfs and not rootforscan:
        for pi in sys.path:
            p = _matchsub(_apat, pi)
            if p:
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:sys.path for sub pattern: sys.path[x] = %s\n" % (
                        __name__, getcaller_linenumber(),
                        str(_apat),
                        ))
                return p

    # 5. search pattern by PATH
    if not _itfs and not rootforscan:
        for pi in expandpath(os.environ.get('PATH', None)):
            p = _matchsub(_apat, pi)
            if p:
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:env:PATH for sub pattern: PATH[x] = %s\n" % (
                        __name__, getcaller_linenumber(),
                        str(_apat),
                        ))
                return p

    # 6. <HOME>/eclipse/eclipse
    if not p:
        _r = gethome()
        _r += os.sep + 'eclipse' + os.sep + 'eclipse'
        if os.path.exists(_r):
            if kargs.get('altpat'):
                _apat = kargs.get('altpat')
                if _apat.startswith('eclipse/plugins/'):
                    _apat = _apat[7:]
            else:
                _apat = pydevd_glob
            _r = os.path.normpath(
                os.path.dirname(os.path.realpath(_r)) + "/plugins/")

        p = _matchsub(_apat, _r)
        if p:
            return sorted(p)[-1]

    # 7. search install directory when installed within the 'dropins' directory
    if not p:
        if os.path.exists(os.path.normpath("../plugins")) and os.path.exists(
                os.path.normpath("../eclipse")):
            _r = glob.glob(os.path.normpath("../plugins/" + _pg))
        p = _matchsub(_apat, _r)
        if p:
            return p


def scan_eclipse_for_pydevd(rootforscan=None, **kargs):
    """Scans file system directory tree of Eclipse for the PyDev plugin 
    and contained 'pydevd' source directory 'pysrc'.

    See PyDev manual for path reference, the default pattern 
    for installation from 'marketplace' is: ::

       eclipse/plugins/org.python.pydev_x.x.x/pysrc/pydevd.py

    In case of drop-in installation ::

       eclipse/dropins/<pydev-dropin-name>/plugins/org.python.pydev_x.x.x/pysrc/pydevd.py

    The matching versions could be varied by glob-expressions.
    The provided parameters match as follows: ::

       rdbgroot := /path/to/eclipse
       rdbgsub  := org.python.pydev_x.x.x/pysrc/pydevd.py

    The glue-hook depends on the type of installation and is determined
    dynamically: ::

       dropin-install       := dropins/<pydev-dropin-name>/plugins
       market-place-install :=  plugins

    The search order is:

    1. **PATH - which eclipse**

       Each separate path is tried with the sub-path pattern,
       first match wins.

    2. **<HOME>/eclipse/eclipse**

       A convention of the author, where the the path is
       a symbolic link to the executable.
       When present, the realpath is evaluated from the link.

    3. **search install directory - dropins**

       When 'pyrdbg' itself is installed as a drop-in within eclipse,
       the search is performed within the current Eclipse release
       only::

          os.path.dirname(__file__) + rdbgsub

       For example::

           eclipse/dropins/pyrdbg

    The first match of containing directory for 'pydevd.py'
    is returned, thus ambiguity in case of multiple occurrences
    has to be avoided.

    The pattern resolution into the PyDev dir is performed
    by the steps: ::

       0. <eclipse-root>/plugins/<rdbgsub>
       1. <eclipse-root>/dropins/<rdbgsub>
       2. <eclipse-root>/dropins/*/plugins/<rdbgsub>

    For additional details refer to the section 
    'Client Stub Lookup for "pydevd"' in the manual.

    Args:

        **rootforscan**:
            Start directory for tree-scan:

            * a single path string
            * search path notation with multiple entries
            * a *list* of paths

            Each path points to an Eclipse installation
            directory.

    kargs:

        **altpat**:

            Alternative pattern, varies 'eclipse_glob',
            and 'pydevd_glob'. ::

               altpat := (<literal> | <glob>)

        **strict**:

            Provided parameter has to match, else error.

            Default is to try all, when supplied params do not
            match default values are checked.

        **version**:

            Provide version to be requested. It is recommended
            to combine this with 'strict'.

            The parameters 'altpat' and 'version' are EXOR. ::

               version=(a,b,c)

               a := [0-9]+
               b := [0-9]+
               c := [0-9]+

    Returns:

        The location of pydevd.py

    Raises:

        AttributeError:

    """

    def _matchsub(_apat, _rfs):
        """
        Args:
            **_apat**:
                Alternate relative file path name pattern with wildcards.
                For provided *glob* and *re* see *expandpath*.
                The scans are: ::

                   0: as IDE-plugin sub directory
                   1: as IDE-dropin sub directory

            **_rfs**:
                Root directory for path match.

        Returns:
            *Success*:  file path

            *Failure*:  *None*

        Raises:
            pass-through

        """
        # plugins
        p = expandpath(_rfs + os.sep + "plugins" + os.path.sep + _apat, **{'wildcards': True})
        if p:  # a match trial
            return p

        # dropins
        p = expandpath(_rfs + os.sep + "dropins" + os.path.sep + "*" + os.path.sep
                       +"plugins" + os.path.sep + _apat, **{'wildcards': True})
        if p:  # a match trial
            return p

#     # 6. <HOME>/eclipse/eclipse
#     if not p:
#         _r = gethome()
#         _r += os.sep + 'eclipse' + os.sep + 'eclipse'
#         if os.path.exists(_r):
#             if kargs.get('altpat'):
#                 _apat = kargs.get('altpat')
#                 if _apat.startswith('eclipse/plugins/'):
#                     _apat = _apat[7:]
#             else:
#                 _apat = pydevd_glob
#             _r = os.path.normpath(
#                 os.path.dirname(os.path.realpath(_r)) + "/plugins/")
# 
#         p = _matchsub(_apat, _r)
#         if p:
#             return sorted(p)[-1]
# 
#     # 7. search install directory when installed within the 'dropins' directory
#     if not p:
#         if os.path.exists(os.path.normpath("../plugins")) and os.path.exists(
#                 os.path.normpath("../eclipse")):
#             _r = glob.glob(os.path.normpath("../plugins/" + _pg))
#         p = _matchsub(_apat, _r)
#         if p:
#             return p


import rdbg.pydevrdc
