# -*- coding: utf-8 -*-
"""The session controller for the automation of cross-process remote debugging."""
from __future__ import absolute_import
from filesysobjects.pathtools import expandpath

import os
import sys

from sourceinfo.fileinfo import getcaller_linenumber

import rdbg.shared
from rdbg import PyRdbgError, callfromPyDevDbg


__author__ = 'Arno-Can Uestuensoez'
__license__ = "Artistic-License-2.0 + Forced-Fairplay-Constraints"
__copyright__ = "Copyright (C) 2010-2019 Arno-Can Uestuensoez" \
                "@Ingenieurbuero Arno-Can Uestuensoez"
__version__ = '0.2.21'
__uuid__ = '7f464182-b723-4e6e-869d-80a16b9cfaca'

__docformat__ = "restructuredtext en"



class RdbgRdcError(PyRdbgError):
    """Remote debug client base exception class."""
    pass


class RdbgRdcLoadError(RdbgRdcError):
    """Failed load of 'pydevd.py'."""

    def __init__(self):
        super(RdbgRdcError, self).__init__("\n" + \
            "\n" + \
            "Cannot load module 'pydevd':\n" + \
            "\n" + \
            "   Set PYTHONPATH, or use more call parameters.\n" + \
            "   See PyDev manual for path reference:\n" + \
            "     'eclipse/plugins/org.python.pydev_x.x.x/pysrc/pydevd.py'\n" + \
            "\n"
        )


class RdbgRdcServerError(RdbgRdcError):
    """Debug server of PyDev is not running - cannot be reached."""

    def __init__(self):
        super(RdbgRdcError, self).__init__("\n" + \
            "\n" + \
            "Cannot connect to debug server, start it using Eclipse menu:\n" + \
            "\n" + \
            "   In Debug-Perspective: 'PyDev -> Start Debug Server'\n" + \
            "\n"
        )


class PyDevRDC(object):
    """Provides automation for remote debugging of external Python
    process with PyDev and others providing the debugging protocol
    of *PyDev.Debugger* or short *pydevd*.

    This particularly provides a prepared environment for
    cross-process debugging by 'pydevd'.
    """

    #: Controls whether to be initialized.
    _initok = False

    defaultargs = {}

    #: Final default configuration arguments for a debug controller, see pydevd.settrace.
    dgbargs_default = {
        'host': 'localhost',  # host(None)
        'stdoutToServer': True,  # stdoutToServer(False)
        'stderrToServer': True,  # stderrToServer(False)
        'port': 5678,  # port(5678)
        'suspend': False,  # suspend(True) :
        'trace_only_current_thread': True,  # trace_only_current_thread(False)
        # 'ignore' == overwrite_prev_trace(False)
        'overwrite_prev_trace': False,
        'patch_multiprocessing': True,  # patch_multiprocessing(False),
    }

    #: Subpath glob pattern of PyDev plugin.
    pydevd_glob = "org.python.pydev[._][^0-9]*[.0-9]*.[0-9]*.[0-9]*[.0-9]*/pysrc/pydevd.py"

    #: Subpath glob pattern for any eclipse based PyDev installation.
    eclipse_glob = rdbg.shared._rdbgroot_default

    def __init__(self, rdbgsession={}, **kargs):
        """Create a control stub for a remote debugger.
        Search and load 'pydevd' for cross-process remote debugging.

        Args:
            rdbgsession:
                The session data as returned from *checkandremove_rdbgoptions()*.

            **kargs:

                **fpname**:

                    The file path name of the main file for the current
                    process, default:=callname.

                **label**:

                    An optional label for identifying the currrent instance.
                    The label could be provided as debugging flag. ::

                       label := <name>

                    default:=callname.

                **noargv**:

                    Suppresses argv processing completely for the previous
                    arguments. The testflags are still processed.

                **remotedebug**:

                    Switches remote debugging support On/Off. Dependent of
                    this parameter an internal call of startRemoteDebugExt
                    is performed and a new instance initialized.

                **rdbgtestflags**:

                    List of flags to force specific behaviour - mostly faulty -
                    in order to test the module itself. So, **do not use**
                    these if you do not know what these actually do.

                    These partially fail, but provide a sufficient
                    part of the control flow for the aspect of interest.

                        *ignorePydevd*:

                            Debugging the initial bootstrap of an simulated
                            external process. Ignores present loaded debug
                            support. For debugging of the debug support.

                        *ignorePydevdSysPath*:

                            Debugging the initial bootstrap of an simulated
                            external process. Ignores load by current 'sys.path'.
                            For debugging of the debug support.

                        *ignorePydevdCallParam*:

                            Debugging the initial bootstrap of an simulated
                            external process. Ignores current parameter for
                            load. For debugging of the debug support.

        Returns:
            Creates a proxy instance.

        Raises:
            *RdbgRdcError*

            pass-through
        """
        global pydevd

        if rdbg.my_dbg_self > 0:
            res = '{'
            for k, v in rdbgsession.items():
                res += '\n %-15s: %s,' % (str(k), str(v))
            res += '\n}'
            if rdbg.my_dbg_self:
                sys.stderr.write(
                    "RDBG:%s:%d:rdbgsession = %s\n" % (__name__,getcaller_linenumber(),
                    res))

        self._itfp = self._itfs = self._itfcp = False

        # set initial defaults for the *pydevd* API
        self.dbgargs = self.dgbargs_default

        #
        # fetch current CLI values from rdbg.checkrdbg as defaults
        if rdbgsession.get('rdbg'):  # _rdbgsrv
            h, p = rdbgsession.get('rdbg').split(':')
            if h:
                self.dbgargs['host'] = h
            if p:
                self.dbgargs['port'] = int(p)

        self.eclipse_glob = rdbgsession.get('rdbgroot', rdbg.shared._rdbgroot_default)
        self.pydevd_glob = rdbgsession.get('rdbgsub', rdbg.shared._rdbgsub_default)
        self.pydevd_path = rdbgsession.get('rdbgpath', '')
        self.pydev_srcmap = rdbgsession.get('rdbgsrcmap', [])
        self.pydev_label = rdbgsession.get('rdbglabel', rdbg.shared._rdbglabel_default)
        self.pydev_testflags = rdbgsession.get('rdbgtestflags', [])
        self.pydevd = rdbgsession.get('pydevd', None)

        for tf in self.pydev_testflags:
            if tf == 'ignorePydevd':
                self._itfp = True
            elif tf == 'ignorePydevdSysPath':
                self._itfs = True
            elif tf == 'ignorePydevdCallParam':
                self._itfcp = True

        if rdbg.my_dbg_self > 2:
            _lnum = getcaller_linenumber()
            if self.pydev_srcmap:
                sys.stderr.write(
                    "RDBG:%s:%s:pydevd parameters:srcmap = [\n" % (__name__, _lnum)
                )
                for m in self.pydev_srcmap:
                    sys.stderr.write(
                        "RDBG:%s:%s:pydevd parameters:   %s,\n" % (__name__, _lnum, str(m))
                    )
                sys.stderr.write(
                    "RDBG:%s:%s:pydevd parameters:]\n" % (__name__, _lnum)
                )
            else:
                sys.stderr.write(
                    "RDBG:%s:%s:pydevd parameters:srcmap = []\n" % (__name__, _lnum)
                )

#             sys.stderr.write(
#                 "RDBG:%s:%s:pydevd parameters\n" % (
#                     __name__, _lnum) +
#                 "RDBG:%s:%d:srcmap=%s\n" % (__name__, getcaller_linenumber(), str(self.pydev_srcmap))
#             )

        # check whether running in pydevd.py
        if not self._itfp and sys.modules.get('pydevd'):  # already loaded
            if rdbg.my_dbg_self > 2:
                _lnum = getcaller_linenumber()
                sys.stderr.write(
                    "RDBG:%s:%d:Already running in pydevd\n" % (
                    __name__, _lnum) +
                    "RDBG:%s:%d:use loaded pydevd.py\n" % (
                    __name__, _lnum)
                )

            self.runningInPyDevDbg = True

        else:  # has to be loaded, do the initial first-time work
            if rdbg.my_dbg_self > 2:
                sys.stderr.write(
                    "RDBG:%s:%d:search and load pydevd.py\n" % (
                    __name__, getcaller_linenumber()))

                sys.stderr.write(
                    "RDBG:%s:%d:pydevd_path = %s\n" % (
                    __name__, getcaller_linenumber(), str(self.pydevd_path)))

            if self.pydevd_path:  # expand path
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:filesysobjects.pathtools.expandpath()\n" % (
                        __name__, getcaller_linenumber()))

                if type(self.pydevd_path) in (list, tuple,):
                    self.pydevd_path = expandpath(*self.pydevd_path)

                else:
                    self.pydevd_path = expandpath(self.pydevd_path)
                
                if len(self.pydevd_path) > 1:
                    # resulting from the last option
                    self.pydevd_path = sorted(self.pydevd_path)[-1]

                elif len(self.pydevd_path) == 1:
                    # the only one option
                    self.pydevd_path = self.pydevd_path[0]

            if not self.pydevd_path:  # scan path rooot and sub-path
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:search:0. checkrdbg.scan_for_pydevd(%s, altpat=%s)\n" % (
                        __name__, getcaller_linenumber(),
                        str(self.eclipse_glob), 
                        str(self.pydevd_glob),
                        ))

                self.pydevd_path = scan_for_pydevd(self.eclipse_glob, altpat=self.pydevd_glob)
                if self.pydevd_path:
                    self.pydevd_path = sorted(self.pydevd_path)[-1]
            else:
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:pydevd_path = %s\n" % (
                        __name__, getcaller_linenumber(), str(self.pydevd_path)))

            if not self.pydevd_path:  # scane path rooot and sub-path
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:search:1. checkrdbg.scan_for_pydevd(altpat=%s)\n" % (
                        __name__, getcaller_linenumber(),
                        str(self.pydevd_glob),
                        ))

                self.pydevd_path = scan_for_pydevd(altpat=self.pydevd_glob)
                if self.pydevd_path:
                    self.pydevd_path = sorted(self.pydevd_path)[-1]

            if not self.pydevd_path:
                # try PYTHONPATH of package rdbg itself
                _p = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:search:2. check rdbg package path: %s\n" % (
                        __name__, getcaller_linenumber(),
                        str(_p), 
                        ))

                if os.path.exists(_p + os.sep + "pydevd"):
                    if rdbg.my_dbg_self > 2:
                        sys.stderr.write(
                            "RDBG:%s:%d:os.path.dirname(%s)\n" % (
                            __name__, getcaller_linenumber(),
                            str(__file__), 
                            ))
    
                    self.pydevd_path = os.path.normpath(_p)
            
            if self.pydevd_path and os.path.isfile(self.pydevd_path):
                if rdbg.my_dbg_self > 2:
                    sys.stderr.write(
                        "RDBG:%s:%d:os.path.dirname(%s)\n" % (
                        __name__, getcaller_linenumber(),
                        str(self.pydevd_path), 
                        ))

                self.pydevd_path = os.path.dirname(self.pydevd_path)

            if not self.pydevd_path:
                _head = "RDBG:"+ __name__ +":"+ str(getcaller_linenumber()) + ":" 
                raise RdbgRdcError(
                    (
                        '\n' +
                         _head + "Load of module 'pydevd' failed, possible parameters:\n" +
                         _head + "\n" +
                         _head + "   sys.argv                            = %s\n" +
                         _head + "   rdbgroot / --rdbg-root / RDBGROOT   = %s\n" +
                         _head + "   rdbgsub  / --rdbg-sub  / RDBGSUB    = %s\n" +
                         _head + "   rdbgpath / --rdbg-path / RDBGPATH   = %s\n" +
                         _head + "\n" +
                         _head + "ERROR:Cannot find 'pydevd' - cannot get 'rdbgpath' for stub module 'pydevd.py'\n" +
                         _head + "\n" +
                         _head + "INFO: 0. Check your PYTHONPATH if installed 'pydevd' via 'pip install pydevd'.\n" +
                         _head + "INFO: 1. Check path for local copy of directory 'pysrc' from PyDev subdirectory \n" +
                         _head + "INFO: 2. Check default path for local Eclipse subdirectory 'pysrc' from PyDev\n" +
                         _head + "INFO: 3. Check your RDBGPATH or use call parameter '--rdbg-path'.\n" +
                         _head + "INFO: 4. Check your RDBGROOT or use call parameter '--rdbg-root'.\n" +
                         _head + "INFO: 5. Check your RDBGSUB or use call parameter '--rdbg-sub'.\n" +
                         _head + "INFO: 6. Install 'pydevd' or subdirectory 'pysrc' from PyDev\n" +
                         _head + "\n"
                     ) % (
                         str(rdbg.shared.argv_org),
                         str(self.eclipse_glob),
                         str(self.pydevd_glob),
                         str(self.pydevd_path),                         
                     ) 
                )

            try:
                # REMARK: pydevd prevents a second, even from a different location
                #   so debugging of pydevd with another release seems not possible
                sys.path.insert(0, self.pydevd_path)
                import pydevd  # @UnresolvedImport #@UnusedImport
                self.pydevd = pydevd

                self.runningInPyDevDbg = True

            except Exception as e:
                # self.runningInPyDevDbg = False
                raise RdbgRdcError("Load of 'pydevd' failed:" +
                    str(self.pydevd_path) + "\n" + str(e))

            sys.path.pop(0)

        if rdbg.my_dbg_self > 2:
            sys.stderr.write(
                "RDBG:%s:%d:start_debug\n" % (
                __name__, getcaller_linenumber()))

        # starts remote debug immediately
        self.remotedebug = kargs.get("remotedebug", True)

        # ignore errors
        self.ignore = kargs.get("ignore", False)

        if self.remotedebug:
            self.startdebug()
        self._initok = True

    def isactive(self):
        """Checks current debug session state.

        Args:
            *None*.

        Returns:
            *True*:  Active session.

            *False*: Deactivated session.

        Raises:
            pass-through
            
        """
        return callfromPyDevDbg or self.pydevd.connected

    def isattached(self):
        """Checks whether started from within PyDev or as standalone program attached to/by rdbg.
        Check whether *pydevd* is already preloaded, if not decides to be attached.

        Args:
            *None*

        Returns:
            *True*:   Running an attached debug session, not initiated by 'rdgb'.
            
            *False*:  Session was started from within *PyDev*.

        Raises:
            pass-through
            
        """
        return not callfromPyDevDbg and self.pydevd.connected

    def ispydevcalled(self):
        """Checks whether started from within PyDev.
        Check whether *pydevd* is already preloaded, if not decides to be attached.

        Args:
            *None*

        Returns:
            *True*:   Started by *PyDev* e.g. from within *Eclipse*.
            
            *False*:  Session created from external, e.g. via PyDev-Debug-Server.

        Raises:
            pass-through
            
        """
        return callfromPyDevDbg

    def setdebug_params(self, **kargs):
        """Sets the parameters for debug. This overwrites the values set by
        the *rdbgsession* parameter or the initialization. The parameters
        take effect when set before the start of a debug session, or after
        restart of the remote debug session.

        Args:
            kargs:
                **host**:

                    Host name where the debug server is running.

                    From **pdevd.py**:

                        The user may specify another host, if the debug server
                        is not in the same machine (default is the local host).

                **ignore**:

                    When set to False success is mandatory, else
                    an exception it raised.
                    Set this to True/default, when in production
                    systems.

                **port**:

                    Port the debug server is listening on.
                    From **pdevd.py**:

                        Specifies which port to use for communicating with the
                        server (note that the server must be started in the
                        same port).

                    **Note**: the default is hard-coded *5678*

                **remotedebug**:

                    Switches remote debugging support On/Off.

                **stderrToServer**:

                    Sets whether *stderr* is directed to the debug server.
                    From **pdevd.py**:

                        When this is true, the *stderr* is passed to the debug server
                        so that they are printed in its console and not in this
                        process console.

                **stdoutToServer**:

                    Sets whether *stdout* is directed to the debug server.
                    From **pdevd.py**:

                        When this is true, the *stdout* is passed to the debug server.

                **suspend**:

                    If set to True, stops immediately after *settrace()* call,
                    else at next valid break-condition.
                    From **pdevd.py**:

                        Whether a breakpoint should be emulated as soon as
                        this function is called.

                **trace_only_current_thread**:

                    From **pdevd.py**:

                        Determines if only the current thread will be traced or all
                        future threads will also have the tracing enabled.

        Returns:
            The location of pydevd.py

        Raises:
            *AttributeError*

        """
        self.ignore = kargs.get('ignore', True)
        self.remotedebug = kargs.get('remotedebug', True)

        for k, v in kargs.items():
            if k == 'host':
                self.dbgargs['host'] = v
            elif k == 'stdoutToServer':
                self.dbgargs['stdoutToServer'] = v
            elif k == 'stderrToServer':
                self.dbgargs['stderrToServer'] = v
            elif k == 'port':
                self.dbgargs['port'] = v
            elif k == 'suspend':
                self.dbgargs['suspend'] = v
            elif k == 'trace_only_current_thread':
                self.dbgargs['trace_only_current_thread'] = v
            elif k == 'overwrite_prev_trace':
                self.dbgargs['overwrite_prev_trace'] = v
            elif k == 'patch_multiprocessing':
                self.dbgargs['patch_multiprocessing'] = v

    def startdebug(self, **kargs):
        """Starts the remote debugging session for PyDev.

        Args:
            *ignore*:

                Suppress exception, return *False* instead.
                See *setdebug_params* for default.

        Returns:
            *True*:  is debugging

            *False*: debugging is deactivated

        Raises:
            *RdbgRdcError*

            pass-through
            
        """
        _ignore = kargs.get('ignore', self.ignore)

        if not self.remotedebug:
            sys.stderr.write("RDBG:%s:%d:remote debug is deactivated\n" %(__name__, getcaller_linenumber(),))
            return False

        if rdbg.my_dbg_self > 3:
            sys.stderr.write("RDBG:%s:%d:debug starting\n" %(__name__, getcaller_linenumber(),))

        # check already running under control of a debug session
        if callfromPyDevDbg and 'ignorePydevd' not in self.pydev_testflags:
            if rdbg.my_dbg_self:
                sys.stderr.write("RDBG:%s:%d:Application started within a PyDev debug session, no attach\n"  %(__name__, getcaller_linenumber()))
            return True

        # loaded by attached process for remote debug server
        if self.runningInPyDevDbg:
            if rdbg.my_dbg_self > 1:
                sys.stderr.write("RDBG:%s:%d:dbgargs=%s"  %(__name__, getcaller_linenumber(), str(self.dbgargs) + "\n"))

            try:
                # OK-ref: pydevd.settrace(
                #     host=None,port=5678,
                #     stdoutToServer=False,
                #     stderrToServer=False,
                #     suspend=False,
                #     trace_only_current_thread=True
                #     )

                if rdbg.my_dbg_self > 3:
                    ln = getcaller_linenumber()
                    sys.stderr.write("RDBG:%s:%d:pydevd.settrace\n" %(__name__, ln))
                    sys.stderr.write("RDBG:%s:%d:host                      = %s\n" %(__name__, ln, str(self.dbgargs['host'])))
                    sys.stderr.write("RDBG:%s:%d:port                      = %s\n" %(__name__, ln, str(self.dbgargs['port'])))
                    sys.stderr.write("RDBG:%s:%d:stdoutToServer            = %s\n" %(__name__, ln, str(self.dbgargs['stdoutToServer'])))
                    sys.stderr.write("RDBG:%s:%d:stderrToServer            = %s\n" %(__name__, ln, str(self.dbgargs['stderrToServer'])))
                    sys.stderr.write("RDBG:%s:%d:suspend                   = %s\n" %(__name__, ln, str(self.dbgargs['suspend'])))
                    sys.stderr.write("RDBG:%s:%d:trace_only_current_thread = %s\n" %(__name__, ln, str(self.dbgargs['trace_only_current_thread'])))

                self.pydevd.settrace(
                    host=self.dbgargs['host'],
                    port=self.dbgargs['port'],
                    stdoutToServer=self.dbgargs['stdoutToServer'],
                    stderrToServer=self.dbgargs['stderrToServer'],
                    suspend=self.dbgargs['suspend'],
                    trace_only_current_thread=self.dbgargs[
                        'trace_only_current_thread']
                    )

                if rdbg.my_dbg_self > 0:
                    sys.stderr.write("RDBG:%s:%d:pydevd.settrace() => OK\n" %(__name__, getcaller_linenumber(),))

            except Exception as e:
                if _ignore:
                    sys.stderr.write(
                        "Cannot start debug:pydevd.settrace(): pydevd=" + str(self.pydevd) +
                        " - " + str(e)
                    )
                    return False

                raise RdbgRdcError(
                    "Cannot start debug:pydevd.settrace(): pydevd=" + str(self.pydevd) +
                    " - " + str(e)
                )

        else:
            if _ignore:
                sys.stderr.write(
                    "ERROR:Requires init:self.runningInPyDevDbg=" +
                    str(self.runningInPyDevDbg))
                return False

            raise RdbgRdcError(
                "ERROR:Requires init:self.runningInPyDevDbg=" +
                str(self.runningInPyDevDbg))

        if rdbg.my_dbg_self > 0:
            _lnum = getcaller_linenumber()
            sys.stderr.write(
                "RDBG:%s:%d:remote status\n" % (__name__, _lnum,) +
                "RDBG:%s:%d:isactive       = %s\n" % (__name__, _lnum, str(self.isactive())) +
                "RDBG:%s:%d:attached       = %s\n" % (__name__, _lnum, str(self.isattached())) +
                "RDBG:%s:%d:ispydevcalled  = %s\n" % (__name__, _lnum, str(self.ispydevcalled()))
            )
        return True

    def stopdebug(self, **kargs):
        """Stops the remote debugging for *PyDev*.

        Current version does not support a *startdebug()* after *stopdebug()*.

        Args:
            **ignore**:

                Suppress exception, return False instead.
                See 'setdebug_params' for default.

        Returns:
            *True*:  stopped

            *False*: cannot stop

        Raises:
            *RdbgRdcError*

            pass-through
        """
        _ignore = kargs.get('ignore', self.ignore)

        try:
            if self.runningInPyDevDbg:
                self.pydevd.stoptrace()
        except Exception as e:  # @UnusedVariable
            if _ignore:
                sys.stderr.write("ERROR:Cannot stop:" + str(e))
                return False
            raise RdbgRdcError("ERROR:Cannot stop:" + str(e))

        return True

    def setfork(self):
        """Prepares debugging after fork.
        """
        try:
            if self.runningInPyDevDbg:
                self.pydevd.settrace_forked()
        except Exception as e:  # @UnusedVariable
            if self.ignore:
                sys.stderr.write("ERROR:Cannot set fork:" + str(e))
                return
            raise RdbgRdcError("ERROR:Cannot set fork:" + str(e))
        pass

    def __str__(self):
        """Prints the current remote debug parameters.
        """
        return \
            "\nPyDevRDC.host                      = " + str(self.dbgargs['host']) + \
            "\nPyDevRDC.stdoutToServer            = " + \
            str(self.dbgargs['stdoutToServer']) + \
            "\nPyDevRDC.stderrToServer            = " + \
            str(self.dbgargs['stderrToServer']) + \
            "\nPyDevRDC.port                      = " + str(self.dbgargs['port']) + \
            "\nPyDevRDC.suspend                   = " + str(self.dbgargs['suspend']) + \
            "\nPyDevRDC.trace_only_current_thread = " + \
            str(self.dbgargs['trace_only_current_thread']) + \
            "\nPyDevRDC.overwrite_prev_trace      = " + \
            str(self.dbgargs['overwrite_prev_trace']) + \
             "\nPyDevRDC.patch_multiprocessing     = " + \
            str(self.dbgargs['patch_multiprocessing'])

    def __repr__(self):
        """Prints the current representation of remote debug parameters.
        """
        return "{" + \
            "'host': " + str(self.host) + \
            ", 'stdoutToServer': " + str(self.stdoutToServer) + \
            ", 'stderrToServer': " + str(self.stderrToServer) + \
            ", 'port': " + str(self.port) + \
            ", 'suspend': " + str(self.suspend) + \
            ", 'trace_only_current_thread': " + \
            str(self.trace_only_current_thread) + \
            ", 'overwrite_prev_trace': " + str(self.overwrite_prev_trace) + \
            ", 'patch_multiprocessing': " + str(self.patch_multiprocessing) + \
            "}"

def getdefaultproxy():
    """Gets the shared default remote debug controller proxy.

    Args:
        *None*.

    Returns:

        Default proxy instance, the shared static instance of
        *rdbg.pydevrdc.PyDevRDC*.

        The value is *None* when not yet initialized.

    Raises:

        *None*

    """
    return PYDEVD

def setdefaultproxy(dproxy):
    """Sets the shared default remote debug controller proxy.

    Args:

        **dproxy**:
            The new default proxy. requires an instance of type
            *rdbg.pydevrdc.PyDevRDC*.

    Returns:
        The old proxy. The value is *None*, when this was not yet
        initialized.

    Raises:

        *None*
    """
    global PYDEVD

    _o = PYDEVD
    PYDEVD = dproxy
    return _o

def init_rdbg(rdbgsession={}, **kargs):
    """Create a shared control stub for remote debugging.
    Search and load 'pydevd.py' and 'pydev_file_utils.py'
    for cross-process and inter-host remote debugging.

    The shared instance *PYDEVD* of the current process
    is a singleton.

    Args:
        **rdbgsession*:
            The session data as returned from *checkandremove_rdbgoptions()*.

        kargs:

            **debug**:

                Debug RDBG itself, developer output.

            **fpname**:

                The file path name of the main file for the current
                process, default:=callname.

            **label**:

                An optional label for identifying the currrent instance.
                The label could be provided as debugging flag. ::
                
                   label := <name>

            **noargv**:

                Suppresses argv processing completely for the previous
                arguments. The testflags are still processed.

            **remotedebug**:

                Switches remote debugging support On/Off. Dependent of
                this parameter an internal call of startRemoteDebugExt
                is performed and a new instance initialized.

            **shared**:
                Creates a shared instance PYDEVD of class 'PyDevRDC'.
                This is the default controller for *pyrdbg*. ::

                   shared := (True|False)

                   True :  create and return reference to PYDEVD
                   False:  return a private instance

                default := True

            **testflags**:

                Flags to force specific behaviour - mostly faulty -
                in order to test the module itself. So, **do not use**
                these if you do not know what these actually do.

                These partially fail, but provide a sufficient
                part of the control flow for the aspect of interest.

                    *ignorePydevd*:

                        Debugging the initial bootstrap of an simulated
                        external process. Ignores present loaded debug
                        support. For debugging of the debug support.

                    *ignorePydevdSysPath*:

                        Debugging the initial bootstrap of an simulated
                        external process. Ignores load by current 'sys.path'.
                        For debugging of the debug support.

                    *ignorePydevdCallParam*:

                        Debugging the initial bootstrap of an simulated
                        external process. Ignores current parameter for
                        load. For debugging of the debug support.

    Returns:
        Creates a proxy instance.

    Raises:
        pass-through
        
    """
    _shared = kargs.get('shared', True)

    if _shared and (rdbg.shared.PYDEVD and rdbg.shared.PYDEVD._initok):
        if rdbg.my_dbg_self > 1:
            sys.stderr.write(
                "RDBG:%s:%d:already_present-using_it:rdbg.pydevrdc.PYDEVD\n" %(
                    __name__,
                    getcaller_linenumber(),
                    )
                )
#FIXME:        return
#4TEST:
        return rdbg.shared.PYDEVD

    if _shared and not rdbg.shared.PYDEVD:
        if rdbg.my_dbg_self > 1:
            sys.stderr.write(
                "RDBG:%s:%d:create:rdbg.pydevrdc.PYDEVD\n" %(
                    __name__,                
                    getcaller_linenumber(),
                    )
                )

    _PYDEVD = PyDevRDC(rdbgsession, **kargs)
    if _shared:
        rdbg.shared.PYDEVD = _PYDEVD

    return _PYDEVD


from rdbg.checkrdbg import scan_for_pydevd
