# -*- coding: utf-8 -*-
#
# The basic minimalistic call interface for initial bootstrap.
#
from __future__ import absolute_import
import sys

import rdbg

from rdbg import my_dbg_self
from rdbg import my_verb_self

__author__ = 'Arno-Can Uestuensoez'
__license__ = "Artistic-License-2.0 + Forced-Fairplay-Constraints"
__copyright__ = "Copyright (C) 2010-2019 Arno-Can Uestuensoez" \
                " @Ingenieurbuero Arno-Can Uestuensoez"
__version__ = '0.2.22'
__uuid__ = '7f464182-b723-4e6e-869d-80a16b9cfaca'

__docformat__ = "restructuredtext en"


def start_remote_debug(argv=None, **kw):
    """Checks command line for injected options, when activated loads, 
    initializes, and starts the remote debug session.

    Provides the most basic call parameters, while does not require for the
    standard default cases any parameter.

    Args:
        For detailed description refer to the manual.

        **argv**:
            Alternative argv,

        kargs:
            **appname**:
                Label, default is name of the calling application.

            **debugrdbg**:
                Prints debug traces for rdbg itself.

            **enableautoload**:
                Enables the complete headless auto load and proceed.

            **enableautopath**:
                Sets alternate path for 'rdbg-auto.conf' / 'rdbg-auto.json'.

            **forceload**:
                By default *rdbg* is loaded only when the option
                '--rdbg' is present. The flag 'forceload' disables
                the check.

            **host**:
                Host location of the running remote-debug-server instance.

            **path**:
                An explicit path into the 'pysrc' directory of 'PyDev'.

            **port**:
                The configured listening port of the remote-debug-server.

            **rootpath**:
                Path into a selected Eclipse package directory.

            **srcmap**:
                A map of source files to executables. The paths map the
                top directories of the source files to the top directories
                of the executables.

            **subpath**:
                Sub-directory of PyDev release. This could be a *glob*
                and/or *re*.

            **verbrdbg**:
                Prints verbose display for rdbg itself.

    Environment:
        The current release of *RDBG* at the time of writing requires 
        the following libraries::

            jsondata
            multiconf
            platformids
            sourceinfo
            yapyutils

    Returns:
        *True*:  started

        *False*: error condition occurred

    Raises:
        pass-through

    """
    global my_dbg_self
    global my_verb_self
    
    if argv == None:
        argv = sys.argv

    try:
#        rdbg.my_dbg_self = kw.pop('debugrdbg')
        my_dbg_self = kw.pop('debugrdbg')
    except KeyError:
        pass

    try:
#        rdbg.my_verb_self = kw.pop('verbrdbg')
        my_verb_self = kw.pop('verbrdbg')
    except KeyError:
        pass

#FIXME:

    try:
#        if rdbg.my_dbg_self:
        if my_dbg_self:
            sys.stderr.write(
                "RDBG:" + __name__ + ":start_remote_debug:argv=%s\n" % str(argv)
                )
    except Exception as e:
        pass
    
#     if rdbg.my_dbg_self:
#         sys.stderr.write(
#             "RDBG:" + __name__ + ":start_remote_debug:argv=%s\n" % str(argv)
#             )
#    if rdbg.my_verb_self:
    if my_verb_self:
        sys.stdout.write(
            "VERB:" + __name__ + ":start_remote_debug:argv=%s\n" % str(argv)
            )

    _forceload = kw.get('forceload')
    _enableautoload = kw.get('enableautoload')

    if "--rdbg-self" in sys.argv:
        # standalone option with optional following values
        my_dbg_self += 1
        sys.stdout.write(
            "VERB:" + __name__ + ":raised level of remote debug\n"
            )

    # check rdbg in early fetch, thus requires 'hand-made' pre-parsing with small footprint 
    _cli = False
    if "--rdbg" in sys.argv:
        # standalone option with optional following values
        _cli = True

    else:
        # is an option with included parameters by assignment
        for i in sys.argv:
            if i.startswith('--rdbg=') or i in ('--rdbg-h', '--rdbg-help',):
                _cli = True
                break

        # else:
        #    # has no enabling option for '--rdbg'
        #    _cli = False

    # check permitted combinations
    if _cli and _enableautoload:
        raise rdbg.PyRdbgError("PyRDBG command line options and 'enableautoload' are mutual exclusive")
    if _forceload and _enableautoload:
        raise rdbg.PyRdbgError("'forceload' and 'enableautoload' are mutual exclusive")

    # TODO: _tmp = sys.path.pop(0)  # avoid name -collision
    _status = False

    import sourceinfo.objectinfo
    import sourceinfo.fileinfo
    import rdbg.checkrdbg
    
#     from sourceinfo.objectinfo import getcaller_package_name
#     from sourceinfo.fileinfo import getcaller_filepathname
#     from rdbg.checkrdbg import checkandremove_rdbgoptions

#    if rdbg.my_dbg_self:
    if my_dbg_self:
        if _forceload:
            sys.stderr.write(
                "RDBG:" + __name__ + ":hardcoded forceload from:" + str(sourceinfo.fileinfo.getcaller_filepathname(2)) + "\n"
                )

        elif _enableautoload:
            sys.stderr.write(
                "RDBG:" + __name__ + ":hardcoded enableautoload from:" + str(sourceinfo.fileinfo.getcaller_filepathname(2)) + "\n"
                )

        sys.stderr.write(
            "RDBG:" + __name__ + ":start_remote_debug:load\n"
            )

    try:
        label=kw.pop('appname')
    except KeyError:
        label=sourceinfo.objectinfo.getcaller_package_name(2)

#    if rdbg.my_dbg_self > 2:
    if my_dbg_self > 2:
        sys.stderr.write('RDBG:call param kw = %s\n' % (str(kw)))

    _rdbg_param = rdbg.checkrdbg.checkandremove_rdbgoptions(label=label,**kw)
#    if rdbg.my_dbg_self > 2:
    if my_dbg_self > 2:
        sys.stderr.write('RDBG:label      = %s\n' % (str(label)))
        sys.stderr.write('RDBG:rdbg_param = %s\n' % (str(_rdbg_param)))

    if _rdbg_param['rdbgthis'] or _forceload:
        
#        if rdbg.my_dbg_self:
        if my_dbg_self:
            sys.stderr.write('RDBG:initialize rdbg\n')
            
        from rdbg.pydevrdc import init_rdbg  # activate remote debug stub call
        _rdbgd = init_rdbg(_rdbg_param, shared=True)
        if _rdbgd:
            if rdbg.my_dbg_self:
                sys.stderr.write('RDBG:init OK, start debug..\n')
            # start debugging here...
            _status = _rdbgd.startdebug()
    
    # TODO: sys.path.insert(0, _tmp)  # restore for following other libraries
#    if rdbg.my_dbg_self:
    if my_dbg_self:
        sys.stderr.write('RDBG:start_remote_debug:status=%s\n' % (str(_status)))

    return _status

